/**
 * @file
 *
 * This file contains function prototypes for the operations of the
 * <code>BinaryTreeNode</code> class.
 */

#ifndef INCLUSION_GUARD_BINARY_TREE_NODE
#define INCLUSION_GUARD_BINARY_TREE_NODE

#include <stdbool.h>

/**
 * This type represents the <code>BinaryTreeNode</code>
 * class. Instances of this class represent immutable nodes in a
 * binary tree. Arbitrary data may be associated with a node.
 */
typedef struct BinaryTreeNode_s BinaryTreeNode_t;

/**
 * Creates an instance of the <code>BinaryTreeNode</code> class.
 *
 * @param[in] pData
 *   the client-defined data of this node, or <code>NULL</code>.
 * @param[in] pLeftNode
 *   the left node of this node, or <code>NULL</code>.
 * @param[in] pRightNode
 *   the right node of this node, or <code>NULL</code>.
 * @param[out] ppThis
 *   a pointer to the variable which shall hold the instance of the
 *   <code>BinaryTreeNode</code> class. Must not be <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
bool BinaryTreeNode_Create(void* pData,
                           BinaryTreeNode_t* pLeftNode,
                           BinaryTreeNode_t* pRightNode,
                           BinaryTreeNode_t** ppThis);

/**
 * Destroys this instance of the <code>BinaryTreeNode</code>
 * class. The children of this node will not be recursively destroyed
 * as a result of calling this operation. If the client-defined data
 * associated with this node has been initialized to point to memory
 * allocated at runtime, this memory must be manually deallocated
 * before calling this operation. This operation always succeeds.
 *
 * @param[in] pThis
 *   an instance of the <code>BinaryTreeNode</code> class. May be
 *   <code>NULL</code>.
 */
void BinaryTreeNode_Destroy(BinaryTreeNode_t* pThis);

/**
 * Returns the client-defined data associated with this node, or
 * <code>NULL</code> if there is no such data.
 *
 * @param[in] pThis
 *   a pointer to the instance data of this
 *   <code>BinaryTreeNode</code> object. Must not be
 *   <code>NULL</code>.
 * @param[out] ppData
 *   a pointer to the variable which shall hold the client-defined
 *   data. Must not be <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
bool BinaryTreeNode_Data(BinaryTreeNode_t* pThis,
                         void** ppData);

/**
 * Returns the left node of this node, or <code>NULL</code> if there
 * is no such node.
 *
 * @param[in] pThis
 *   a pointer to the instance data of this
 *   <code>BinaryTreeNode</code> object. Must not be
 *   <code>NULL</code>.
 * @param[out] ppLeftNode
 *   a pointer to the variable which shall hold the reference to the
 *   left node. Must not be <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
bool BinaryTreeNode_LeftNode(BinaryTreeNode_t* pThis,
                             BinaryTreeNode_t** ppLeftNode);

/**
 * Returns the right node of this node, or <code>NULL</code> if there
 * is no such node.
 *
 * @param[in] pThis
 *   a pointer to the instance data of this
 *   <code>BinaryTreeNode</code> object. Must not be
 *   <code>NULL</code>.
 * @param[out] ppRightNode
 *   a pointer to the variable which shall hold the reference to the
 *   right node. Must not be <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
bool BinaryTreeNode_RightNode(BinaryTreeNode_t* pThis,
                              BinaryTreeNode_t** ppRightNode);

#endif // INCLUSION_GUARD_BINARY_TREE_NODE
