/**
 * @file
 *
 * This file contains a class, <code>DefaultBinaryOperatorNode</code>,
 * which is the default implementation of the
 * <code>BinaryOperatorNode</code> interface.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdarg.h>
#include "Node.h"
#include "BinaryOperatorNode.h"
#include "BinaryOperator.h"
#include "DefaultBinaryOperatorNode.h"

/**
 * This structure represents the instance data of
 * <code>DefaultBinaryOperatorNode</code> objects.
 */
typedef struct
{
  /**
   * The dispatch table used by this instance.
   */
  const BinaryOperatorNode_DispatchTable_t* pDispatchTable;

  /**
   * The binary operator used by this object.
   */
  BinaryOperator_t operator;

  /**
   * The node representing the left operand. This member is never
   * <code>NULL</code>.
   */
  Node_t* pLeftOperand;

  /**
   * The node representing the right operand. This member is never
   * <code>NULL</code>.
   */
  Node_t* pRightOperand;
} DefaultBinaryOperatorNode_InstanceData_t;

static const BinaryOperatorNode_DispatchTable_t
  gBinaryOperatorNodeDispatchTable;

bool DefaultBinaryOperatorNode_Create(BinaryOperator_t operator,
                                      Node_t* pLeftOperand,
                                      Node_t* pRightOperand,
                                      BinaryOperatorNode_t** ppResult)
{
  DefaultBinaryOperatorNode_InstanceData_t* pInstanceData = NULL;

  bool result =
    (pLeftOperand != NULL) &&
    (pRightOperand != NULL) &&
    (ppResult != NULL);

  if (result)
  {
    pInstanceData =
      malloc(sizeof(DefaultBinaryOperatorNode_InstanceData_t));
    result = pInstanceData != NULL;
  }

  if (result)
  {
    pInstanceData->pDispatchTable = &gBinaryOperatorNodeDispatchTable;
    pInstanceData->operator = operator;
    pInstanceData->pLeftOperand = pLeftOperand;
    pInstanceData->pRightOperand = pRightOperand;
    *ppResult = (BinaryOperatorNode_t*)pInstanceData;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

static void DefaultBinaryOperatorNode_Destroy(
  BinaryOperatorNode_t* pBinaryOperatorNode)
{
  if (pBinaryOperatorNode != NULL)
  {
    DefaultBinaryOperatorNode_InstanceData_t* pThis =
      (DefaultBinaryOperatorNode_InstanceData_t*)pBinaryOperatorNode;

    Node_Destroy(pThis->pLeftOperand);
    Node_Destroy(pThis->pRightOperand);

    free(pBinaryOperatorNode);
  }
}

static bool DefaultBinaryOperatorNode_IsConstant(
  BinaryOperatorNode_t* pBinaryOperatorNode, bool* pResult)
{
  bool result = (pBinaryOperatorNode != NULL) && (pResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    (DefaultBinaryOperatorNode_InstanceData_t*)pBinaryOperatorNode;
  bool isLeftOperandConstant = false;
  bool isRightOperandConstant = false;

  if (result)
  {
    result = Node_IsConstant(pThis->pLeftOperand,
                             &isLeftOperandConstant);
  }

  if (result)
  {
    result = Node_IsConstant(pThis->pRightOperand,
                             &isRightOperandConstant);
  }

  if (result)
  {
    *pResult = isLeftOperandConstant && isRightOperandConstant;
  }
  else if (pResult != NULL)
  {
    *pResult = false;
  }

  return result;
}

static bool DefaultBinaryOperatorNode_Operator(
  BinaryOperatorNode_t* pBinaryOperatorNode,
  BinaryOperator_t* pResult)
{
  bool result = (pBinaryOperatorNode != NULL) && (pResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    (DefaultBinaryOperatorNode_InstanceData_t*)pBinaryOperatorNode;

  if (result)
  {
    *pResult = pThis->operator;
  }
  else if (pResult != NULL)
  {
    *pResult = BinaryOperator_UNDEFINED;
  }

  return result;
}

static bool DefaultBinaryOperatorNode_LeftOperand(
  BinaryOperatorNode_t* pBinaryOperatorNode,
  Node_t** ppResult)
{
  bool result = (pBinaryOperatorNode != NULL) && (ppResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    (DefaultBinaryOperatorNode_InstanceData_t*)pBinaryOperatorNode;

  if (result)
  {
    *ppResult = pThis->pLeftOperand;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

static bool DefaultBinaryOperatorNode_RightOperand(
  BinaryOperatorNode_t* pBinaryOperatorNode,
  Node_t** ppResult)
{
  bool result = (pBinaryOperatorNode != NULL) && (ppResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    (DefaultBinaryOperatorNode_InstanceData_t*)pBinaryOperatorNode;

  if (result)
  {
    *ppResult = pThis->pRightOperand;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

/**
 * This is the dispatch table for the <code>BinaryOperatorNode</code>
 * interface implemented by this class.
 */
static const BinaryOperatorNode_DispatchTable_t
  gBinaryOperatorNodeDispatchTable =
{
  DefaultBinaryOperatorNode_Destroy,
  DefaultBinaryOperatorNode_IsConstant,
  DefaultBinaryOperatorNode_Operator,
  DefaultBinaryOperatorNode_LeftOperand,
  DefaultBinaryOperatorNode_RightOperand
};
