/**
 * @file
 *
 * This file contains a class, <code>DefaultUnaryOperatorNode</code>,
 * which is the default implementation of the
 * <code>UnaryOperatorNode</code> interface.
 */

#include <stdbool.h>
#include <stdlib.h>
#include "Node.h"
#include "UnaryOperatorNode.h"
#include "UnaryOperator.h"
#include "DefaultUnaryOperatorNode.h"

/**
 * This structure represents the instance data of
 * <code>DefaultUnaryOperatorNode</code> objects.
 */
typedef struct DefaultUnaryOperatorNode_InstanceData_s
{
  /**
   * The dispatch table used by this instance.
   */
  const UnaryOperatorNode_DispatchTable_t* pDispatchTable;

  /**
   * The unary operator used by this object.
   */
  UnaryOperator_t operator;

  /**
   * The node representing the sole operand. This member is never
   * <code>NULL</code>.
   */
  Node_t* pOperand;
} DefaultUnaryOperatorNode_InstanceData_t;

static const UnaryOperatorNode_DispatchTable_t
  gUnaryOperatorNodeDispatchTable;

bool DefaultUnaryOperatorNode_Create(UnaryOperator_t operator,
                                     Node_t* pOperand,
                                     UnaryOperatorNode_t** ppResult)
{
  DefaultUnaryOperatorNode_InstanceData_t* pInstanceData = NULL;

  bool result = (pOperand != NULL) && (ppResult != NULL);

  if (result)
  {
    pInstanceData =
      malloc(sizeof(DefaultUnaryOperatorNode_InstanceData_t));
    result = pInstanceData != NULL;
  }

  if (result)
  {
    pInstanceData->pDispatchTable = &gUnaryOperatorNodeDispatchTable;
    pInstanceData->operator = operator;
    pInstanceData->pOperand = pOperand;
    *ppResult = (UnaryOperatorNode_t*)pInstanceData;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

static void DefaultUnaryOperatorNode_Destroy(
  UnaryOperatorNode_t* pUnaryOperatorNode)
{
  if (pUnaryOperatorNode != NULL)
  {
    DefaultUnaryOperatorNode_InstanceData_t* pThis =
      (DefaultUnaryOperatorNode_InstanceData_t*)pUnaryOperatorNode;

    Node_Destroy(pThis->pOperand);

    free(pUnaryOperatorNode);
  }
}

static bool DefaultUnaryOperatorNode_IsConstant(
  UnaryOperatorNode_t* pUnaryOperatorNode, bool* pResult)
{
  bool result = (pUnaryOperatorNode != NULL) && (pResult != NULL);
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    (DefaultUnaryOperatorNode_InstanceData_t*)pUnaryOperatorNode;
  bool isOperandConstant = false;

  if (result)
  {
    result = Node_IsConstant(pThis->pOperand, &isOperandConstant);
  }

  if (result)
  {
    *pResult = isOperandConstant;
  }
  else if (pResult != NULL)
  {
    *pResult = false;
  }

  return result;
}

static bool DefaultUnaryOperatorNode_Operator(
  UnaryOperatorNode_t* pUnaryOperatorNode, UnaryOperator_t* pResult)
{
  bool result = (pUnaryOperatorNode != NULL) && (pResult != NULL);
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    (DefaultUnaryOperatorNode_InstanceData_t*)pUnaryOperatorNode;

  if (result)
  {
    *pResult = pThis->operator;
  }
  else if (pResult != NULL)
  {
    *pResult = UnaryOperator_UNDEFINED;
  }

  return result;
}

static bool DefaultUnaryOperatorNode_Operand(
  UnaryOperatorNode_t* pUnaryOperatorNode,
  Node_t** ppResult)
{
  bool result = (pUnaryOperatorNode != NULL) && (ppResult != NULL);
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    (DefaultUnaryOperatorNode_InstanceData_t*)pUnaryOperatorNode;

  if (result)
  {
    *ppResult = pThis->pOperand;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

/**
 * This is the dispatch table for the <code>UnaryOperatorNode</code>
 * interface implemented by this class.
 */
static const UnaryOperatorNode_DispatchTable_t
  gUnaryOperatorNodeDispatchTable =
{
  DefaultUnaryOperatorNode_Destroy,
  DefaultUnaryOperatorNode_IsConstant,
  DefaultUnaryOperatorNode_Operator,
  DefaultUnaryOperatorNode_Operand
};
