/**
 * @file
 *
 * This file contains an interface representing a identifier operand
 * node in a syntax tree. This tree represents an arithmetic
 * expression. A identifier operand node holds an identifier character
 * (in the expression "3 / x", "x" is a identifier). This interface
 * extends <code>Node</code>.
 */

#ifndef INCLUSION_GUARD_IDENTIFIER_OPERAND_NODE
#define INCLUSION_GUARD_IDENTIFIER_OPERAND_NODE

#include <stdbool.h>
#include <wchar.h>
#include "Node.h"

/* Convenience macros for the operations inherited from the Node
 * interface:
 */
#define IdentifierOperandNode_Destroy(pIdentifierOperandNode)           \
  if ((pIdentifierOperandNode) != NULL)                                 \
  {                                                                     \
    (pIdentifierOperandNode)->pDispatchTable->Destroy(                  \
      pIdentifierOperandNode);                                          \
  }
#define IdentifierOperandNode_IsConstant(pIdentifierOperandNode,        \
                                         pResult)                       \
  (pIdentifierOperandNode)->pDispatchTable->IsConstant(                 \
    (pIdentifierOperandNode),                                           \
    (pResult))

// Convenience macros for the operations defined in this interface:

/**
 * Returns the identifier character of this node.
 *
 * @param[in] pIdentifierOperandNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] ppResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the identifier character of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define IdentifierOperandNode_IdentifierCharacter(                      \
  pIdentifierOperandNode, ppResult)                                     \
  (pIdentifierOperandNode)->pDispatchTable->IdentifierCharacter(        \
    (pIdentifierOperandNode),                                           \
    (ppResult))

struct IdentifierOperandNode_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct IdentifierOperandNode_DispatchTable_s* pDispatchTable;
} IdentifierOperandNode_t;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>IdentifierOperandNode_t</code> type to match their own. (Only
 * single interface inheritance is supported.)
 */
typedef struct IdentifierOperandNode_DispatchTable_s
{
  // Operations inherited from the Node interface:
  void (*Destroy)(IdentifierOperandNode_t* pIdentifierOperandNode);
  bool (*IsConstant)(IdentifierOperandNode_t* pIdentifierOperandNode, 
                     bool* pResult);

  // Operations defined in the IdentifierOperandNode interface:
  bool (*IdentifierCharacter)(
    IdentifierOperandNode_t* pIdentifierOperandNode,
    wchar_t* pResult);
} IdentifierOperandNode_DispatchTable_t;

#endif // INCLUSION_GUARD_IDENTIFIER_OPERAND_NODE
