/**
 * @file
 *
 * This file contains an interface representing a literal operand node
 * in a syntax tree. This tree represents an arithmetic expression. A
 * literal operand node holds a constant (in the expression "3 / x",
 * "3" is a literal). This interface extends <code>Node</code>.
 */

#ifndef INCLUSION_GUARD_LITERAL_OPERAND_NODE
#define INCLUSION_GUARD_LITERAL_OPERAND_NODE

#include <stdbool.h>
#include "Node.h"

/* Convenience macros for the operations inherited from the Node
 * interface:
 */
#define LiteralOperandNode_Destroy(pLiteralOperandNode)                 \
  if ((pLiteralOperandNode) != NULL)                                    \
  {                                                                     \
    (pLiteralOperandNode)->pDispatchTable->Destroy(pLiteralOperandNode);\
  }
#define LiteralOperandNode_IsConstant(pLiteralOperandNode, pResult)     \
  (pLiteralOperandNode)->pDispatchTable->IsConstant(                    \
    (pLiteralOperandNode), (pResult))

// Convenience macros for the operations defined in this interface:

/**
 * Returns the constant of this node.
 *
 * @param[in] pLiteralOperandNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] ppResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the constant of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define LiteralOperandNode_Constant(pLiteralOperandNode, ppResult)      \
  (pLiteralOperandNode)->pDispatchTable->Constant((pLiteralOperandNode),\
                                                  (ppResult))

struct LiteralOperandNode_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct LiteralOperandNode_DispatchTable_s* pDispatchTable;
} LiteralOperandNode_t;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>LiteralOperandNode_t</code> type to match their own. (Only
 * single interface inheritance is supported.)
 */
typedef struct LiteralOperandNode_DispatchTable_s
{
  // Operations inherited from the Node interface:
  void (*Destroy)(LiteralOperandNode_t* pLiteralOperandNode);
  bool (*IsConstant)(LiteralOperandNode_t* pLiteralOperandNode,
                     bool* pResult);

  // Operations defined in the LiteralOperandNode interface:
  bool (*Constant)(LiteralOperandNode_t* pLiteralOperandNode,
                   double* pResult);
} LiteralOperandNode_DispatchTable_t;

#endif // INCLUSION_GUARD_LITERAL_OPERAND_NODE
