/**
 * @file
 *
 * This file contains a program that tests the classes present in this
 * directory. In particular, it makes sure that dynamic dispatch works
 * as intended.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include "Node.h"
#include "BinaryOperatorNode.h"
#include "DefaultBinaryOperatorNode.h"
#include "UnaryOperatorNode.h"
#include "DefaultUnaryOperatorNode.h"
#include "LiteralOperandNode.h"
#include "DefaultLiteralOperandNode.h"
#include "IdentifierOperandNode.h"
#include "DefaultIdentifierOperandNode.h"

/**
 * Creates a test hierarchy of nodes, corresponding to this arithmetic
 * expression: ((-x - (6 * -3)) / z) + 2
 *
 * This is the resulting hierarchy:
 *
 * <code>
 *  1. BinaryOperatorNode[operator=+]
 *  2.   BinaryOperatorNode[operator=/]
 *  3.     BinaryOperatorNode[operator=-]
 *  4.       UnaryOperatorNode[operator=-]
 *  5.         IdentifierOperandNode[identifiercharacter=y]
 *  6.       BinaryOperatorNode[operator=*]
 *  7.         LiteralOperandNode[constant=6]
 *  8.         UnaryOperatorNode[operator=-]
 *  9.           LiteralOperandNode[constant=3]
 * 10.     IdentifierOperandNode[identifierCharacter=z]
 * 11.   LiteralOperandNode[constant=2]
 * </code>
 *
 * @param[out] ppRootNode
 *   a pointer to the variable which shall hold the instance of the
 *   root node (number 1, in the list above). Must not be
 *   <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
static bool createNodes(BinaryOperatorNode_t** ppRootNode)
{
  bool result = (ppRootNode != NULL);

  /* The number suffixes of the variable names correspond to the
   * numbers given to the nodes in the list above.
   */
  BinaryOperatorNode_t* pNode1 = NULL;
  BinaryOperatorNode_t* pNode2 = NULL;
  BinaryOperatorNode_t* pNode3 = NULL;
  UnaryOperatorNode_t* pNode4 = NULL;
  IdentifierOperandNode_t* pNode5 = NULL;
  BinaryOperatorNode_t* pNode6 = NULL;
  LiteralOperandNode_t* pNode7 = NULL;
  UnaryOperatorNode_t* pNode8 = NULL;
  LiteralOperandNode_t* pNode9 = NULL;
  IdentifierOperandNode_t* pNode10 = NULL;
  LiteralOperandNode_t* pNode11 = NULL;

  if (result)
  {
    result = DefaultLiteralOperandNode_Create(3, &pNode9);
  }

  if (result)
  {
    result = DefaultUnaryOperatorNode_Create(UnaryOperator_NEGATION, 
                                             (Node_t*)pNode9,
                                             &pNode8);
  }

  if (result)
  {
    result = DefaultLiteralOperandNode_Create(6, &pNode7);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_MULTIPLICATION,
                                              (Node_t*)pNode7,
                                              (Node_t*)pNode8,
                                              &pNode6);
  }

  if (result)
  {
    result = DefaultIdentifierOperandNode_Create(L'y', &pNode5);
  }

  if (result)
  {
    result = DefaultUnaryOperatorNode_Create(UnaryOperator_NEGATION,
                                             (Node_t*)pNode5,
                                             &pNode4);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_SUBTRACTION,
                                              (Node_t*)pNode4,
                                              (Node_t*)pNode6,
                                              &pNode3);
  }

  if (result)
  {
    result = DefaultIdentifierOperandNode_Create(L'z', &pNode10);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_DIVISION,
                                              (Node_t*)pNode3,
                                              (Node_t*)pNode10,
                                              &pNode2);
  }

  if (result)
  {
    result = DefaultLiteralOperandNode_Create(2, &pNode11);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_ADDITION,
                                              (Node_t*)pNode2,
                                              (Node_t*)pNode11,
                                              &pNode1);
  }

  if (result)
  {
    *ppRootNode = pNode1;
  }
  else
  {
    *ppRootNode = NULL;
  }

  return result;
}

/**
 * Prints some information on the root node given as a parameter to
 * this function, as well as some related to its operands.
 *
 * @param[in] pRootNode
 *   a Node reference to the root node of the tree.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
bool experiment(BinaryOperatorNode_t* pRootNode)
{
  Node_t* pLeftOperand = NULL;
  Node_t* pRightOperand = NULL;
  bool rootIsConstant = false;
  bool leftOperandIsConstant = false;
  bool rightOperandIsConstant = false;
  bool result = pRootNode != NULL;

  if (result)
  {
    result = BinaryOperatorNode_LeftOperand(pRootNode,
                                            &pLeftOperand);
  }

  if (result)
  {
    result = BinaryOperatorNode_RightOperand(pRootNode,
                                             &pRightOperand);
  }

  if (result)
  {
    result = Node_IsConstant(pLeftOperand,
                             &leftOperandIsConstant);
  }
  
  if (result)
  {
    result = Node_IsConstant(pRightOperand,
                             &rightOperandIsConstant);
  }

  if (result)
  {
    printf("Root is constant: %d.\n", rootIsConstant);
    printf("Left operand is constant: %d.\n", leftOperandIsConstant);
    printf("Right operand is constant: %d.\n", rightOperandIsConstant);
  }

  return result;
}

/**
 * The main function of this program. It builds a sample arithmetic
 * syntax tree, and prints some information on it using late
 * binding. Essentially, this is a testbed for the object-oriented
 * framework developed in the directory in which this file resides.
 *
 * @return
 *   <code>0</code> if successful, a different value otherwise.
 */
int main()
{
  bool result = true;
  BinaryOperatorNode_t* pRootNode = NULL;

  result = createNodes(&pRootNode);

  if (result)
  {
    result = experiment(pRootNode);
  }

  if (!result)
  {
    printf("Something did not go as intended.\n");
  }

  BinaryOperatorNode_Destroy(pRootNode);

  return result ? 0 : 1;
}
