/**
 * @file
 *
 * This file contains an interface, <code>BinaryOperatorNode</code>,
 * representing a binary operator node in a syntax tree. This tree
 * represents an arithmetic expression. A binary operator takes two
 * operands (in the expression "3 / x", the division operator operates
 * on the literal operand "3" and the identifier operand "x"). This
 * interface extends <code>Node</code>.
 */

#ifndef INCLUSION_GUARD_BINARY_OPERATOR_NODE
#define INCLUSION_GUARD_BINARY_OPERATOR_NODE

#include <stdbool.h>
#include "Fundamental.h"
#include "Node.h"
#include "BinaryOperator.h"

/**
 * The name of this interface.
 */
#define BINARY_OPERATOR_NODE_NAME                                       \
 (L"se.polberger.components.BinaryOperatorNode")

/* Convenience macros for the operations inherited from the
 * Fundamental interface:
 */
#define BinaryOperatorNode_SwitchInterface(pBinaryOperatorNode,         \
                                           pInterfaceName,              \
                                           ppResult)                    \
  (pBinaryOperatorNode)->pDispatchTable->SwitchInterface(               \
    (pBinaryOperatorNode),                                              \
    (pInterfaceName),                                                   \
    (ppResult));
#define BinaryOperatorNode_AddReference(pBinaryOperatorNode)            \
  if ((pBinaryOperatorNode) != NULL)                                    \
  {                                                                     \
    (pBinaryOperatorNode)->pDispatchTable->AddReference(                \
      pBinaryOperatorNode);                                             \
  }
#define BinaryOperatorNode_RemoveReference(pBinaryOperatorNode)         \
  if ((pBinaryOperatorNode) != NULL)                                    \
  {                                                                     \
    (pBinaryOperatorNode)->pDispatchTable->RemoveReference(             \
      pBinaryOperatorNode);                                             \
  }

/* Convenience macros for the operations inherited from the Node
 * interface:
 */
#define BinaryOperatorNode_IsConstant(pBinaryOperatorNode, pResult)     \
  (pBinaryOperatorNode)->pDispatchTable->IsConstant(pBinaryOperatorNode,\
                                                    pResult)
#define BinaryOperatorNode_PrintDebugInformation(pBinaryOperatorNode,   \
                                                 startPosition,         \
                                                 indentationSize)       \
  (pBinaryOperatorNode)->pDispatchTable->PrintDebugInformation(         \
    (pBinaryOperatorNode), (startPosition), (indentationSize))

// Convenience macros for the operations defined in this interface:

/**
 * Returns the binary operator of this node.
 *
 * @param[in] pBinaryOperatorNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] pResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the binary operator of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define BinaryOperatorNode_Operator(pBinaryOperatorNode, pResult)       \
  (pBinaryOperatorNode)->pDispatchTable->Operator((pBinaryOperatorNode),\
                                                  (pResult))

/**
 * Returns the left operand of this node.
 *
 * @param[in] pBinaryOperatorNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] ppResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the left operand of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define BinaryOperatorNode_LeftOperand(pBinaryOperatorNode, ppResult)   \
  (pBinaryOperatorNode)->pDispatchTable->LeftOperand(                   \
    (pBinaryOperatorNode),                                              \
    (ppResult))

/**
 * Returns the right operand of this node.
 *
 * @param[in] pBinaryOperatorNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] ppResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the right operand of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define BinaryOperatorNode_RightOperand(pBinaryOperatorNode, ppResult)  \
  (pBinaryOperatorNode)->pDispatchTable->RightOperand(                  \
    (pBinaryOperatorNode),                                              \
    (ppResult))

struct BinaryOperatorNode_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct BinaryOperatorNode_DispatchTable_s* pDispatchTable;
} BinaryOperatorNode_t;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>BinaryOperatorNode_t</code> type to match their own. (Only
 * single interface inheritance is supported.)
 */
typedef struct BinaryOperatorNode_DispatchTable_s
{
  // Operations inherited from the Fundamental interface:
  bool (*SwitchInterface)(BinaryOperatorNode_t* pBinaryOperatorNode,
                          wchar_t* pInterfaceName,
                          Fundamental_t** ppResult);
  void (*AddReference)(BinaryOperatorNode_t* pBinaryOperatorNode);
  void (*RemoveReference)(BinaryOperatorNode_t* pBinaryOperatorNode);

  // Operations inherited from the Node interface:
  bool (*IsConstant)(BinaryOperatorNode_t* pBinaryOperatorNode,
                     bool* pResult);
  bool (*PrintDebugInformation)(
    BinaryOperatorNode_t* pBinaryOperatorNode,
    unsigned int startPosition,
    unsigned int indentationSize);

  // Operations defined in the BinaryOperatorNode interface:
  bool (*Operator)(BinaryOperatorNode_t* pBinaryOperatorNode,
                   BinaryOperator_t* pResult);
  bool (*LeftOperand)(BinaryOperatorNode_t* pBinaryOperatorNode,
                      Node_t** ppResult);
  bool (*RightOperand)(BinaryOperatorNode_t* pBinaryOperatorNode,
                       Node_t** ppResult);
} BinaryOperatorNode_DispatchTable_t;

#endif // INCLUSION_GUARD_BINARY_OPERATOR_NODE
