/**
 * @file
 *
 * This file contains a class, <code>DefaultBinaryOperatorNode</code>,
 * which is the default implementation of the
 * <code>BinaryOperatorNode</code> interface. It also implements the
 * <code>Scriptable</code> interface.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "Fundamental.h"
#include "Node.h"
#include "BinaryOperatorNode.h"
#include "BinaryOperator.h"
#include "DefaultBinaryOperatorNode.h"
#include "Scriptable.h"
#include "ScriptableArgumentType.h"

struct DefaultBinaryOperatorNode_InstanceData_s;

/**
 * This structure represents an interface node, which enables access
 * to instances of the class contained in this file.
 */
typedef struct
{
  /**
   * The dispatch table, or vtable. The type is not specified, as this
   * structure is not used to access the dispatch table (whose type
   * may vary), but to access the instance data.
   */
  const void* pDispatchTable;

  /**
   * The instance data.
   */
  struct DefaultBinaryOperatorNode_InstanceData_s* pInstanceData;
} DefaultBinaryOperatorNode_InterfaceNode_t;

/**
 * This structure represents the instance data of
 * <code>DefaultBinaryOperatorNode</code> objects.
 */
typedef struct DefaultBinaryOperatorNode_InstanceData_s
{
  /**
   * The interface node that makes it possible to access this instance
   * through the <code>BinaryOperatorNode</code> interface (and all
   * interfaces that this interface descends from).
   */
  DefaultBinaryOperatorNode_InterfaceNode_t
    interfaceNodeBinaryOperatorNode;

  /**
   * The interface node that makes it possible to access this instance
   * through the <code>Scriptable</code> interface (and all interfaces
   * that this interface descends from).
   */
  DefaultBinaryOperatorNode_InterfaceNode_t interfaceNodeScriptable;

  /**
   * The reference count of this instance. When this drops to zero,
   * the instance is automatically destroyed.
   */
  unsigned int referenceCount;

  /**
   * The binary operator used by this object.
   */
  BinaryOperator_t operator;

  /**
   * The node representing the left operand. This member is never
   * <code>NULL</code>.
   */
  Node_t* pLeftOperand;

  /**
   * The node representing the right operand. This member is never
   * <code>NULL</code>.
   */
  Node_t* pRightOperand;
} DefaultBinaryOperatorNode_InstanceData_t;

static const BinaryOperatorNode_DispatchTable_t
  gBinaryOperatorNodeDispatchTable;
static const Scriptable_DispatchTable_t gScriptableDispatchTable;

bool DefaultBinaryOperatorNode_Create(BinaryOperator_t operator,
                                      Node_t* pLeftOperand,
                                      Node_t* pRightOperand,
                                      wchar_t* pInterfaceName,
                                      Fundamental_t** ppResult)
{
  DefaultBinaryOperatorNode_InstanceData_t* pInstanceData = NULL;

  bool result =
    (pLeftOperand != NULL) &&
    (pRightOperand != NULL) &&
    (pInterfaceName != NULL) &&
    (ppResult != NULL);

  if (result)
  {
    pInstanceData =
      malloc(sizeof(DefaultBinaryOperatorNode_InstanceData_t));
    result = pInstanceData != NULL;
  }

  if (result)
  {
    pInstanceData->interfaceNodeBinaryOperatorNode.pDispatchTable =
      &gBinaryOperatorNodeDispatchTable;
    pInstanceData->interfaceNodeBinaryOperatorNode.pInstanceData =
      pInstanceData;

    pInstanceData->interfaceNodeScriptable.pDispatchTable =
      &gScriptableDispatchTable;
    pInstanceData->interfaceNodeScriptable.pInstanceData =
      pInstanceData;

    pInstanceData->referenceCount = 1;
    pInstanceData->operator = operator;

    /* We are about to make copies of the operand nodes passed to this
     * constructor that we will keep as part of our instance data. As
     * a result, we need to add references to them.
     */
    Node_AddReference(pLeftOperand);
    pInstanceData->pLeftOperand = pLeftOperand;
    Node_AddReference(pRightOperand);
    pInstanceData->pRightOperand = pRightOperand;

    /* Switch to the desired interface. This operation automatically
     * adds a reference to the reference it returns.
     */
    result = Fundamental_SwitchInterface(
      (Fundamental_t*)&(pInstanceData->interfaceNodeBinaryOperatorNode),
      pInterfaceName,
      ppResult);
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  /* Regardless of whether we have failed or succeeded in switching to
   * the desired interface, we need to remove the reference this
   * constructor has added to the instance (implicitly, by setting the
   * reference count to 1).
   */
  if (pInstanceData != NULL)
  {
    Fundamental_RemoveReference(
      (Fundamental_t*)&(pInstanceData->interfaceNodeBinaryOperatorNode));
  }

  return result;
}

static bool DefaultBinaryOperatorNode_SwitchInterface(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pInterfaceName,
  Fundamental_t** ppResult)
{
  bool result = (pInterfaceNode != NULL) && (pInterfaceName != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if ((wcscmp(pInterfaceName, FUNDAMENTAL_NAME) == 0) ||
        (wcscmp(pInterfaceName, NODE_NAME) == 0) ||
        (wcscmp(pInterfaceName, BINARY_OPERATOR_NODE_NAME) == 0))
    {
      if (ppResult != NULL)
      {
        *ppResult =
          (Fundamental_t*)&(pThis->interfaceNodeBinaryOperatorNode);
      }
    }
    else if (wcscmp(pInterfaceName, SCRIPTABLE_NAME) == 0)
    {
      if (ppResult != NULL)
      {
        *ppResult = (Fundamental_t*)&(pThis->interfaceNodeScriptable);
      }
    }
    else
    {
      result = false;
    }
  }

  if (ppResult != NULL)
  {
    if (result)
    {
      Fundamental_AddReference(*ppResult);
    }
    else
    {
      *ppResult = NULL;
    }
  }

  return result;
}

static void DefaultBinaryOperatorNode_AddReference(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultBinaryOperatorNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;
    pThis->referenceCount++;
  }
}

static void DefaultBinaryOperatorNode_RemoveReference(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultBinaryOperatorNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;

    pThis->referenceCount--;

    if (pThis->referenceCount == 0)
    {
      Node_RemoveReference(pThis->pLeftOperand);
      Node_RemoveReference(pThis->pRightOperand);
      free(pThis);
    }
  }
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_IS_CONSTANT (L"IsConstant")

static bool DefaultBinaryOperatorNode_IsConstant(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  bool* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;
  bool isLeftOperandConstant = false;
  bool isRightOperandConstant = false;

  if (result)
  {
    result = Node_IsConstant(pThis->pLeftOperand,
                             &isLeftOperandConstant);
  }

  if (result)
  {
    result = Node_IsConstant(pThis->pRightOperand,
                             &isRightOperandConstant);
  }

  if (result)
  {
    *pResult = isLeftOperandConstant && isRightOperandConstant;
  }
  else if (pResult != NULL)
  {
    *pResult = false;
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION                    \
  (L"PrintDebugInformation")

static bool DefaultBinaryOperatorNode_PrintDebugInformation(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  unsigned int startPosition,
  unsigned int indentationSize)
{
  bool result = pInterfaceNode != NULL;
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if (startPosition != 0)
    {
      printf("%*s", startPosition, " ");
    }

    printf("DefaultBinaryOperatorNode[operator=%d, referenceCount=%d]\n",
           pThis->operator,
           pThis->referenceCount);

    result = Node_PrintDebugInformation(pThis->pLeftOperand,
                                        startPosition + indentationSize,
                                        indentationSize);
  }

  if (result)
  {
    result = Node_PrintDebugInformation(pThis->pRightOperand,
                                        startPosition + indentationSize,
                                        indentationSize);
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_OPERATOR (L"Operator")

static bool DefaultBinaryOperatorNode_Operator(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  BinaryOperator_t* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    *pResult = pThis->operator;
  }
  else if (pResult != NULL)
  {
    *pResult = BinaryOperator_UNDEFINED;
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_LEFT_OPERAND (L"LeftOperand")

static bool DefaultBinaryOperatorNode_LeftOperand(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  Node_t** ppResult)
{
  bool result = (pInterfaceNode != NULL) && (ppResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    Node_AddReference(pThis->pLeftOperand);
    *ppResult = pThis->pLeftOperand;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_RIGHT_OPERAND (L"RightOperand")

static bool DefaultBinaryOperatorNode_RightOperand(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  Node_t** ppResult)
{
  bool result = (pInterfaceNode != NULL) && (ppResult != NULL);
  DefaultBinaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    Node_AddReference(pThis->pRightOperand);
    *ppResult = pThis->pRightOperand;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

static bool DefaultBinaryOperatorNode_InvokeOperation(
  DefaultBinaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pOperationName,
  ScriptableReturnValue_t* pReturnValue,
  ScriptableArgumentType_t firstArgumentType,
  ...)
{
  bool result =
    (pInterfaceNode != NULL) &&
    (pOperationName != NULL) &&
    (pReturnValue != NULL) &&
    (firstArgumentType != ScriptableArgumentType_INVALID) &&
    (firstArgumentType != ScriptableArgumentType_VOID);
  BinaryOperatorNode_t* pBinaryOperatorNode =
    (BinaryOperatorNode_t*)
    &(pInterfaceNode->pInstanceData->interfaceNodeBinaryOperatorNode);

  va_list arguments;
  va_start(arguments, firstArgumentType);

  if (result)
  {
    if (wcscmp(pOperationName, SCRIPTABLE_OPERATION_IS_CONSTANT) == 0)
    {
      bool isConstant = false;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = BinaryOperatorNode_IsConstant(pBinaryOperatorNode,
                                               &isConstant);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_BOOLEAN;
        pReturnValue->values.booleanValue = isConstant;
      }
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION) == 0)
    {
      ScriptableArgumentType_t lastType = firstArgumentType;
      unsigned int argumentCount = 0;
      unsigned int startPosition = 0;
      unsigned int indentationSize = 0;

      // This operation takes two arguments, both integers.
      result = (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS);

      while (result &&
             (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS))
      {
        argumentCount++;
        result = argumentCount <= 2;

        if (result)
        {
          result = (lastType == ScriptableArgumentType_INTEGER);
        }

        if (result)
        {
          switch (argumentCount)
          {
            case 1:
              startPosition = va_arg(arguments, unsigned int);
              break;

            case 2:
              indentationSize = va_arg(arguments, unsigned int);
              break;

            default:
              result = false;
              break;
          }
        }

        if (result)
        {
          lastType = va_arg(arguments, ScriptableArgumentType_t);
        }
      }

      if (result)
      {
        result =
          BinaryOperatorNode_PrintDebugInformation(pBinaryOperatorNode,
                                                   startPosition,
                                                   indentationSize);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_VOID;
      }
    }
    else if (wcscmp(pOperationName, SCRIPTABLE_OPERATION_OPERATOR) == 0)
    {
      BinaryOperator_t operator = BinaryOperator_UNDEFINED;

      // This operation takes no input arguments.
      result = 
	(firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = BinaryOperatorNode_Operator(pBinaryOperatorNode,
                                             &operator);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_INTEGER;
        pReturnValue->values.integerValue = (unsigned int)operator;
      }
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_LEFT_OPERAND) == 0)
    {
      Node_t* pNode = NULL;
      Scriptable_t* pScriptableNode = NULL;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = BinaryOperatorNode_LeftOperand(pBinaryOperatorNode,
                                                &pNode);
      }

      if (result)
      {
        /* The return value is an object accessed through the Node
         * interface. We can only return this value if the class
         * implementing this interface also supports the Scriptable
         * interface, and we thus need to test for this at runtime.
         */
        result = Node_SwitchInterface(pNode,
                                      SCRIPTABLE_NAME,
                                      (Fundamental_t**)&pScriptableNode);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_SCRIPTABLE;
        pReturnValue->values.pScriptableValue = pScriptableNode;
      }

      Node_RemoveReference(pNode);
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_RIGHT_OPERAND) == 0)
    {
      Node_t* pNode = NULL;
      Scriptable_t* pScriptableNode = NULL;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = BinaryOperatorNode_RightOperand(pBinaryOperatorNode,
                                                 &pNode);
      }

      if (result)
      {
        /* The return value is an object accessed through the Node
         * interface. We can only return this value if the class
         * implementing this interface also supports the Scriptable
         * interface, and we thus need to test for this at runtime.
         */
        result = Node_SwitchInterface(pNode,
                                      SCRIPTABLE_NAME,
                                      (Fundamental_t**)&pScriptableNode);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_SCRIPTABLE;
        pReturnValue->values.pScriptableValue = pScriptableNode;
      }

      Node_RemoveReference(pNode);
    }
    else
    {
      result = false;
    }
  }

  va_end(arguments);

  if ((!result) && (pReturnValue != NULL))
  {
    pReturnValue->type = ScriptableArgumentType_INVALID;
  }

  return result;
}

/**
 * This is the dispatch table for the <code>BinaryOperatorNode</code>
 * interface implemented by this class.
 */
static const BinaryOperatorNode_DispatchTable_t
  gBinaryOperatorNodeDispatchTable =
{
  (bool (*)(BinaryOperatorNode_t*, wchar_t*, Fundamental_t**))
    DefaultBinaryOperatorNode_SwitchInterface,
  (void (*)(BinaryOperatorNode_t*))
    DefaultBinaryOperatorNode_AddReference,
  (void (*)(BinaryOperatorNode_t*))
    DefaultBinaryOperatorNode_RemoveReference,
  (bool (*)(BinaryOperatorNode_t*, bool*))
    DefaultBinaryOperatorNode_IsConstant,
  (bool (*)(BinaryOperatorNode_t*, unsigned int, unsigned int))
    DefaultBinaryOperatorNode_PrintDebugInformation,
  (bool (*)(BinaryOperatorNode_t*, BinaryOperator_t*))
    DefaultBinaryOperatorNode_Operator,
  (bool (*)(BinaryOperatorNode_t*, Node_t**))
    DefaultBinaryOperatorNode_LeftOperand,
  (bool (*)(BinaryOperatorNode_t*, Node_t**))
    DefaultBinaryOperatorNode_RightOperand
};

/**
 * This is the dispatch table for the <code>Scriptable</code>
 * interface implemented by this class.
 */
static const Scriptable_DispatchTable_t gScriptableDispatchTable =
{
  (bool (*)(Scriptable_t*, wchar_t*, Fundamental_t**))
    DefaultBinaryOperatorNode_SwitchInterface,
  (void (*)(Scriptable_t*))
    DefaultBinaryOperatorNode_AddReference,
  (void (*)(Scriptable_t*))
    DefaultBinaryOperatorNode_RemoveReference,
  (bool (*)(Scriptable_t*,
            wchar_t*,
            ScriptableReturnValue_t*,
            ScriptableArgumentType_t,
            ...))
    DefaultBinaryOperatorNode_InvokeOperation
};
