/**
 * @file
 *
 * This file contains a class,
 * <code>DefaultIdentifierOperandNode</code>, which is the default
 * implementation of the <code>IdentifierOperandNode</code>
 * interface. It also implements the <code>Scriptable</code>
 * interface.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <wchar.h>
#include <stdarg.h>
#include "Fundamental.h"
#include "Node.h"
#include "IdentifierOperandNode.h"
#include "DefaultIdentifierOperandNode.h"
#include "Scriptable.h"

struct DefaultIdentifierOperandNode_InstanceData_s;

/**
 * This structure represents an interface node, which enables access
 * to instances of the class contained in this file.
 */
typedef struct
{
  /**
   * The dispatch table, or vtable. The type is not specified, as this
   * structure is not used to access the dispatch table (whose type
   * may vary), but to access the instance data.
   */
  const void* pDispatchTable;

  /**
   * The instance data.
   */
  struct DefaultIdentifierOperandNode_InstanceData_s* pInstanceData;
} DefaultIdentifierOperandNode_InterfaceNode_t;

/**
 * This structure represents the instance data of
 * <code>DefaultIdentifierOperandNode</code> objects.
 */
typedef struct DefaultIdentifierOperandNode_InstanceData_s
{
  /**
   * The interface node that makes it possible to access this instance
   * through the <code>IdentifierOperandNode</code> interface (and all
   * interfaces that this interface descends from).
   */
  DefaultIdentifierOperandNode_InterfaceNode_t
    interfaceNodeIdentifierOperandNode;

  /**
   * The interface node that makes it possible to access this instance
   * through the <code>Scriptable</code> interface (and all interfaces
   * that this interface descends from).
   */
  DefaultIdentifierOperandNode_InterfaceNode_t interfaceNodeScriptable;

  /**
   * The reference count of this instance. When this drops to zero,
   * the instance is automatically destroyed.
   */
  unsigned int referenceCount;

  /**
   * The identifier character of this node.
   */
  wchar_t character;
} DefaultIdentifierOperandNode_InstanceData_t;

static const IdentifierOperandNode_DispatchTable_t
  gIdentifierOperandNodeDispatchTable;
static const Scriptable_DispatchTable_t gScriptableDispatchTable;

bool DefaultIdentifierOperandNode_Create(wchar_t character,
                                         wchar_t* pInterfaceName,
                                         Fundamental_t** ppResult)
{
  DefaultIdentifierOperandNode_InstanceData_t* pInstanceData = NULL;

  bool result = (pInterfaceName != NULL) && (ppResult != NULL);

  if (result)
  {
    pInstanceData =
      malloc(sizeof(DefaultIdentifierOperandNode_InstanceData_t));
    result = pInstanceData != NULL;
  }

  if (result)
  {
    pInstanceData->interfaceNodeIdentifierOperandNode.pDispatchTable =
      &gIdentifierOperandNodeDispatchTable;
    pInstanceData->interfaceNodeIdentifierOperandNode.pInstanceData =
      pInstanceData;

    pInstanceData->interfaceNodeScriptable.pDispatchTable =
      &gScriptableDispatchTable;
    pInstanceData->interfaceNodeScriptable.pInstanceData =
      pInstanceData;

    pInstanceData->referenceCount = 1;
    pInstanceData->character = character;

    /* Switch to the desired interface. This operation automatically
     * adds a reference to the reference it returns.
     */
    result = Fundamental_SwitchInterface(
      (Fundamental_t*)&(pInstanceData->interfaceNodeIdentifierOperandNode),
      pInterfaceName,
      ppResult);
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  /* Regardless of whether we have failed or succeeded in switching to
   * the desired interface, we need to remove the reference this
   * constructor has added to the instance (implicitly, by setting the
   * reference count to 1).
   */
  if (pInstanceData != NULL)
  {
    Fundamental_RemoveReference(
      (Fundamental_t*)&(pInstanceData->interfaceNodeIdentifierOperandNode));
  }

  return result;
}

static bool DefaultIdentifierOperandNode_SwitchInterface(
  DefaultIdentifierOperandNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pInterfaceName,
  Fundamental_t** ppResult)
{
  bool result = (pInterfaceNode != NULL) && (pInterfaceName != NULL);
  DefaultIdentifierOperandNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if ((wcscmp(pInterfaceName, FUNDAMENTAL_NAME) == 0) ||
        (wcscmp(pInterfaceName, NODE_NAME) == 0) ||
        (wcscmp(pInterfaceName, IDENTIFIER_OPERAND_NODE_NAME) == 0))
    {
      if (ppResult != NULL)
      {
        *ppResult =
          (Fundamental_t*)&(pThis->interfaceNodeIdentifierOperandNode);
      }
    }
    else if (wcscmp(pInterfaceName, SCRIPTABLE_NAME) == 0)
    {
      if (ppResult != NULL)
      {
        *ppResult = (Fundamental_t*)&(pThis->interfaceNodeScriptable);
      }
    }
    else
    {
      result = false;
    }
  }

  if (ppResult != NULL)
  {
    if (result)
    {
      Fundamental_AddReference(*ppResult);
    }
    else
    {
      *ppResult = NULL;
    }
  }

  return result;
}

static void DefaultIdentifierOperandNode_AddReference(
  DefaultIdentifierOperandNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultIdentifierOperandNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;

    pThis->referenceCount++;
  }
}

static void DefaultIdentifierOperandNode_RemoveReference(
  DefaultIdentifierOperandNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultIdentifierOperandNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;

    pThis->referenceCount--;

    if (pThis->referenceCount == 0)
    {
      free(pThis);
    }
  }
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_IS_CONSTANT (L"IsConstant")

static bool DefaultIdentifierOperandNode_IsConstant(
  DefaultIdentifierOperandNode_InterfaceNode_t* pInterfaceNode,
  bool* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);

  if (pResult != NULL)
  {
    *pResult = false;
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION                    \
  (L"PrintDebugInformation")

static bool DefaultIdentifierOperandNode_PrintDebugInformation(
  DefaultIdentifierOperandNode_InterfaceNode_t* pInterfaceNode,
  unsigned int startPosition,
  unsigned int indentationSize)
{
  bool result = pInterfaceNode != NULL;
  DefaultIdentifierOperandNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if (startPosition != 0)
    {
      printf("%*s", startPosition, " ");
    }

    printf("DefaultIdentifierOperandNode[identifierCharacter=%lc, "
           "referenceCount=%d]\n",
           pThis->character,
           pThis->referenceCount);
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_IDENTIFIER_CHARACTER                       \
  (L"IdentifierCharacter")

static bool DefaultIdentifierOperandNode_IdentifierCharacter(
  DefaultIdentifierOperandNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);
  DefaultIdentifierOperandNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    *pResult = pThis->character;
  }
  else if (pResult != NULL)
  {
    *pResult = L'\0';
  }

  return result;
}

static bool DefaultIdentifierOperandNode_InvokeOperation(
  DefaultIdentifierOperandNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pOperationName,
  ScriptableReturnValue_t* pReturnValue,
  ScriptableArgumentType_t firstArgumentType,
  ...)
{
  bool result =
    (pInterfaceNode != NULL) &&
    (pOperationName != NULL) &&
    (pReturnValue != NULL) &&
    (firstArgumentType != ScriptableArgumentType_VOID);
  IdentifierOperandNode_t* pIdentifierOperandNode =
    (IdentifierOperandNode_t*)
    &(pInterfaceNode->pInstanceData->interfaceNodeIdentifierOperandNode);

  va_list arguments;
  va_start(arguments, firstArgumentType);

  if (result)
  {
    if (wcscmp(pOperationName, SCRIPTABLE_OPERATION_IS_CONSTANT) == 0)
    {
      bool isConstant = false;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = IdentifierOperandNode_IsConstant(pIdentifierOperandNode,
                                                  &isConstant);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_BOOLEAN;
        pReturnValue->values.booleanValue = isConstant;
      }
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION) == 0)
    {
      ScriptableArgumentType_t lastType = firstArgumentType;
      unsigned int argumentCount = 0;
      unsigned int startPosition = 0;
      unsigned int indentationSize = 0;

      // This operation takes two arguments, both integers.
      result = (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS);

      while (result &&
             (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS))
      {
        argumentCount++;
        result = argumentCount <= 2;

        if (result)
        {
          result = (lastType == ScriptableArgumentType_INTEGER);
        }

        if (result)
        {
          switch (argumentCount)
          {
            case 1:
              startPosition = va_arg(arguments, unsigned int);
              break;

            case 2:
              indentationSize = va_arg(arguments, unsigned int);
              break;

            default:
              result = false;
              break;
          }
        }

        if (result)
        {
          lastType = va_arg(arguments, ScriptableArgumentType_t);
        }
      }

      if (result)
      {
        result = IdentifierOperandNode_PrintDebugInformation(
          pIdentifierOperandNode,
          startPosition,
          indentationSize);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_VOID;
      }
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_IDENTIFIER_CHARACTER) == 0)
    {
      wchar_t character = L'\0';

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = IdentifierOperandNode_IdentifierCharacter(
          pIdentifierOperandNode,
          &character);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_CHARACTER;
        pReturnValue->values.characterValue = character;
      }
    }
  }

  va_end(arguments);

  if ((!result) && (pReturnValue != NULL))
  {
    pReturnValue->type = ScriptableArgumentType_INVALID;
  }

  return result;
}

/**
 * This is the dispatch table for the <code>IdentifierOperandNode</code>
 * interface implemented by this class.
 */
static const IdentifierOperandNode_DispatchTable_t
  gIdentifierOperandNodeDispatchTable =
{
  (bool (*)(IdentifierOperandNode_t*, wchar_t*, Fundamental_t**))
    DefaultIdentifierOperandNode_SwitchInterface,
  (void (*)(IdentifierOperandNode_t*))
    DefaultIdentifierOperandNode_AddReference,
  (void (*)(IdentifierOperandNode_t*))
    DefaultIdentifierOperandNode_RemoveReference,
  (bool (*)(IdentifierOperandNode_t*, bool*))
    DefaultIdentifierOperandNode_IsConstant,
  (bool (*)(IdentifierOperandNode_t*, unsigned int, unsigned int))
    DefaultIdentifierOperandNode_PrintDebugInformation,
  (bool (*)(IdentifierOperandNode_t*, wchar_t*))
    DefaultIdentifierOperandNode_IdentifierCharacter
};

/**
 * This is the dispatch table for the <code>Scriptable</code>
 * interface implemented by this class.
 */
static const Scriptable_DispatchTable_t gScriptableDispatchTable =
{
  (bool (*)(Scriptable_t*, wchar_t*, Fundamental_t**))
    DefaultIdentifierOperandNode_SwitchInterface,
  (void (*)(Scriptable_t*))
    DefaultIdentifierOperandNode_AddReference,
  (void (*)(Scriptable_t*))
    DefaultIdentifierOperandNode_RemoveReference,
  (bool (*)(Scriptable_t*,
            wchar_t*,
            ScriptableReturnValue_t*,
            ScriptableArgumentType_t,
            ...))
    DefaultIdentifierOperandNode_InvokeOperation
};
