/**
 * @file
 *
 * This file contains a class, <code>DefaultLiteralOperandNode</code>,
 * which is the default implementation of the
 * <code>LiteralOperandNode</code> interface. It also implements the
 * <code>Scriptable</code> interface.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "Fundamental.h"
#include "Node.h"
#include "LiteralOperandNode.h"
#include "DefaultLiteralOperandNode.h"
#include "Scriptable.h"

struct DefaultLiteralOperandNode_InstanceData_s;

/**
 * This structure represents an interface node, which enables access
 * to instances of the class contained in this file.
 */
typedef struct
{
  /**
   * The dispatch table, or vtable. The type is not specified, as this
   * structure is not used to access the dispatch table (whose type
   * may vary), but to access the instance data.
   */
  const void* pDispatchTable;

  /**
   * The instance data.
   */
  struct DefaultLiteralOperandNode_InstanceData_s* pInstanceData;
} DefaultLiteralOperandNode_InterfaceNode_t;

/**
 * This structure represents the instance data of
 * <code>DefaultLiteralOperandNode</code> objects.
 */
typedef struct DefaultLiteralOperandNode_InstanceData_s
{
  /**
   * The interface node that makes it possible to access this instance
   * through the <code>LiteralOperandNode</code> interface (and all
   * interfaces that this interface descends from).
   */
  DefaultLiteralOperandNode_InterfaceNode_t
    interfaceNodeLiteralOperandNode;

  /**
   * The interface node that makes it possible to access this instance
   * through the <code>Scriptable</code> interface (and all interfaces
   * that this interface descends from).
   */
  DefaultLiteralOperandNode_InterfaceNode_t interfaceNodeScriptable;

  /**
   * The reference count of this instance. When this drops to zero,
   * the instance is automatically destroyed.
   */
  unsigned int referenceCount;

  /**
   * The constant of this node.
   */
  double constant;
} DefaultLiteralOperandNode_InstanceData_t;

static const LiteralOperandNode_DispatchTable_t
  gLiteralOperandNodeDispatchTable;
static const Scriptable_DispatchTable_t gScriptableDispatchTable;

bool DefaultLiteralOperandNode_Create(double constant,
                                      wchar_t* pInterfaceName,
                                      Fundamental_t** ppResult)
{
  DefaultLiteralOperandNode_InstanceData_t* pInstanceData = NULL;

  bool result = (pInterfaceName != NULL) && (ppResult != NULL);

  if (result)
  {
    pInstanceData =
      malloc(sizeof(DefaultLiteralOperandNode_InstanceData_t));
    result = pInstanceData != NULL;
  }

  if (result)
  {
    pInstanceData->interfaceNodeLiteralOperandNode.pDispatchTable =
      &gLiteralOperandNodeDispatchTable;
    pInstanceData->interfaceNodeLiteralOperandNode.pInstanceData =
      pInstanceData;

    pInstanceData->interfaceNodeScriptable.pDispatchTable =
      &gScriptableDispatchTable;
    pInstanceData->interfaceNodeScriptable.pInstanceData =
      pInstanceData;

    pInstanceData->referenceCount = 1;
    pInstanceData->constant = constant;

    /* Switch to the desired interface. This operation automatically
     * adds a reference to the reference it returns.
     */
    result = Fundamental_SwitchInterface(
      (Fundamental_t*)&(pInstanceData->interfaceNodeLiteralOperandNode),
      pInterfaceName,
      ppResult);
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  /* Regardless of whether we have failed or succeeded in switching to
   * the desired interface, we need to remove the reference this
   * constructor has added to the instance (implicitly, by setting the
   * reference count to 1).
   */
  if (pInstanceData != NULL)
  {
    Fundamental_RemoveReference(
      (Fundamental_t*)&(pInstanceData->interfaceNodeLiteralOperandNode));
  }

  return result;
}

static bool DefaultLiteralOperandNode_SwitchInterface(
  DefaultLiteralOperandNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pInterfaceName,
  Fundamental_t** ppResult)
{
  bool result = (pInterfaceNode != NULL) && (pInterfaceName != NULL);
  DefaultLiteralOperandNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if ((wcscmp(pInterfaceName, FUNDAMENTAL_NAME) == 0) ||
        (wcscmp(pInterfaceName, NODE_NAME) == 0) ||
        (wcscmp(pInterfaceName, LITERAL_OPERAND_NODE_NAME) == 0))
    {
      if (ppResult != NULL)
      {
        *ppResult =
          (Fundamental_t*)&(pThis->interfaceNodeLiteralOperandNode);
      }
    }
    else if (wcscmp(pInterfaceName, SCRIPTABLE_NAME) == 0)
    {
      if (ppResult != NULL)
      {
        *ppResult = (Fundamental_t*)&(pThis->interfaceNodeScriptable);
      }
    }
    else
    {
      result = false;
    }
  }

  if (ppResult != NULL)
  {
    if (result)
    {
      Fundamental_AddReference(*ppResult);
    }
    else
    {
      *ppResult = NULL;
    }
  }

  return result;
}

static void DefaultLiteralOperandNode_AddReference(
  DefaultLiteralOperandNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultLiteralOperandNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;

    pThis->referenceCount++;
  }
}

static void DefaultLiteralOperandNode_RemoveReference(
  DefaultLiteralOperandNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultLiteralOperandNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;

    pThis->referenceCount--;

    if (pThis->referenceCount == 0)
    {
      free(pThis);
    }
  }
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_IS_CONSTANT (L"IsConstant")

static bool DefaultLiteralOperandNode_IsConstant(
  DefaultLiteralOperandNode_InterfaceNode_t* pInterfaceNode,
  bool* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);

  if (result)
  {
    *pResult = true;
  }
  else if (pResult != NULL)
  {
    *pResult = false;
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION                    \
  (L"PrintDebugInformation")

static bool DefaultLiteralOperandNode_PrintDebugInformation(
  DefaultLiteralOperandNode_InterfaceNode_t* pInterfaceNode,
  unsigned int startPosition,
  unsigned int indentationSize)
{
  bool result = pInterfaceNode != NULL;
  DefaultLiteralOperandNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if (startPosition != 0)
    {
      printf("%*s", startPosition, " ");
    }

    printf("DefaultLiteralOperandNode[constant=%f, referenceCount=%d]\n",
           pThis->constant,
           pThis->referenceCount);
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_CONSTANT (L"Constant")

static bool DefaultLiteralOperandNode_Constant(
  DefaultLiteralOperandNode_InterfaceNode_t* pInterfaceNode,
  double* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);
  DefaultLiteralOperandNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    *pResult = pThis->constant;
  }
  else if (pResult != NULL)
  {
    *pResult = 0.0;
  }

  return result;
}

static bool DefaultLiteralOperandNode_InvokeOperation(
  DefaultLiteralOperandNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pOperationName,
  ScriptableReturnValue_t* pReturnValue,
  ScriptableArgumentType_t firstArgumentType,
  ...)
{
  bool result =
    (pInterfaceNode != NULL) &&
    (pOperationName != NULL) &&
    (pReturnValue != NULL) &&
    (firstArgumentType != ScriptableArgumentType_VOID);
  LiteralOperandNode_t* pLiteralOperandNode =
    (LiteralOperandNode_t*)
    &(pInterfaceNode->pInstanceData->interfaceNodeLiteralOperandNode);

  va_list arguments;
  va_start(arguments, firstArgumentType);

  if (result)
  {
    if (wcscmp(pOperationName, SCRIPTABLE_OPERATION_IS_CONSTANT) == 0)
    {
      bool isConstant = false;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = LiteralOperandNode_IsConstant(pLiteralOperandNode,
                                               &isConstant);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_BOOLEAN;
        pReturnValue->values.booleanValue = isConstant;
      }
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION) == 0)
    {
      ScriptableArgumentType_t lastType = firstArgumentType;
      unsigned int argumentCount = 0;
      unsigned int startPosition = 0;
      unsigned int indentationSize = 0;

      // This operation takes two arguments, both integers.
      result = (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS);

      while (result &&
             (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS))
      {
        argumentCount++;
        result = argumentCount <= 2;

        if (result)
        {
          result = (lastType == ScriptableArgumentType_INTEGER);
        }

        if (result)
        {
          switch (argumentCount)
          {
            case 1:
              startPosition = va_arg(arguments, unsigned int);
              break;

            case 2:
              indentationSize = va_arg(arguments, unsigned int);
              break;

            default:
              result = false;
              break;
          }
        }

        if (result)
        {
          lastType = va_arg(arguments, ScriptableArgumentType_t);
        }
      }

      if (result)
      {
        result =
          LiteralOperandNode_PrintDebugInformation(pLiteralOperandNode,
                                                   startPosition,
                                                   indentationSize);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_VOID;
      }
    }
    else if (wcscmp(pOperationName, SCRIPTABLE_OPERATION_CONSTANT) == 0)
    {
      double constant = 0.0;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = LiteralOperandNode_Constant(pLiteralOperandNode,
                                             &constant);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_DOUBLE;
        pReturnValue->values.doubleValue = constant;
      }
    }
  }

  va_end(arguments);

  if ((!result) && (pReturnValue != NULL))
  {
    pReturnValue->type = ScriptableArgumentType_INVALID;
  }

  return result;
}

/**
 * This is the dispatch table for the <code>LiteralOperandNode</code>
 * interface implemented by this class.
 */
static const LiteralOperandNode_DispatchTable_t
  gLiteralOperandNodeDispatchTable =
{
  (bool (*)(LiteralOperandNode_t*, wchar_t*, Fundamental_t**))
    DefaultLiteralOperandNode_SwitchInterface,
  (void (*)(LiteralOperandNode_t*))
    DefaultLiteralOperandNode_AddReference,
  (void (*)(LiteralOperandNode_t*))
    DefaultLiteralOperandNode_RemoveReference,
  (bool (*)(LiteralOperandNode_t*, bool*))
    DefaultLiteralOperandNode_IsConstant,
  (bool (*)(LiteralOperandNode_t*, unsigned int, unsigned int))
    DefaultLiteralOperandNode_PrintDebugInformation,
  (bool (*)(LiteralOperandNode_t*, double*))
    DefaultLiteralOperandNode_Constant
};

/**
 * This is the dispatch table for the <code>Scriptable</code>
 * interface implemented by this class.
 */
static const Scriptable_DispatchTable_t gScriptableDispatchTable =
{
  (bool (*)(Scriptable_t*, wchar_t*, Fundamental_t**))
    DefaultLiteralOperandNode_SwitchInterface,
  (void (*)(Scriptable_t*))
    DefaultLiteralOperandNode_AddReference,
  (void (*)(Scriptable_t*))
    DefaultLiteralOperandNode_RemoveReference,
  (bool (*)(Scriptable_t*,
            wchar_t*,
            ScriptableReturnValue_t*,
            ScriptableArgumentType_t,
            ...))
    DefaultLiteralOperandNode_InvokeOperation
};
