/**
 * @file
 *
 * This file contains a class, <code>DefaultUnaryOperatorNode</code>,
 * which is the default implementation of the
 * <code>UnaryOperatorNode</code> interface. It also implements the
 * <code>Scriptable</code> interface.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "Fundamental.h"
#include "Node.h"
#include "UnaryOperatorNode.h"
#include "UnaryOperator.h"
#include "DefaultUnaryOperatorNode.h"
#include "Scriptable.h"
#include "ScriptableArgumentType.h"

struct DefaultUnaryOperatorNode_InstanceData_s;

/**
 * This structure represents an interface node, which enables access
 * to instances of the class contained in this file.
 */
typedef struct
{
  /**
   * The dispatch table, or vtable. The type is not specified, as this
   * structure is not used to access the dispatch table (whose type
   * may vary), but to access the instance data.
   */
  const void* pDispatchTable;

  /**
   * The instance data.
   */
  struct DefaultUnaryOperatorNode_InstanceData_s* pInstanceData;
} DefaultUnaryOperatorNode_InterfaceNode_t;

/**
 * This structure represents the instance data of
 * <code>DefaultUnaryOperatorNode</code> objects.
 */
typedef struct DefaultUnaryOperatorNode_InstanceData_s
{
  /**
   * The interface node that makes it possible to access this instance
   * through the <code>UnaryOperatorNode</code> interface (and all
   * interfaces that this interface descends from).
   */
  DefaultUnaryOperatorNode_InterfaceNode_t
    interfaceNodeUnaryOperatorNode;

  /**
   * The interface node that makes it possible to access this instance
   * through the <code>Scriptable</code> interface (and all interfaces
   * that this interface descends from).
   */
  DefaultUnaryOperatorNode_InterfaceNode_t interfaceNodeScriptable;

  /**
   * The reference count of this instance. When this drops to zero,
   * the instance is automatically destroyed.
   */
  unsigned int referenceCount;

  /**
   * The unary operator used by this object.
   */
  UnaryOperator_t operator;

  /**
   * The node representing the sole operand. This member is never
   * <code>NULL</code>.
   */
  Node_t* pOperand;
} DefaultUnaryOperatorNode_InstanceData_t;

static const UnaryOperatorNode_DispatchTable_t
  gUnaryOperatorNodeDispatchTable;
static const Scriptable_DispatchTable_t gScriptableDispatchTable;

bool DefaultUnaryOperatorNode_Create(UnaryOperator_t operator,
                                     Node_t* pOperand,
                                     wchar_t* pInterfaceName,
                                     Fundamental_t** ppResult)
{
  DefaultUnaryOperatorNode_InstanceData_t* pInstanceData = NULL;

  bool result =
    (pOperand != NULL) &&
    (pInterfaceName != NULL) &&
    (ppResult != NULL);

  if (result)
  {
    pInstanceData =
      malloc(sizeof(DefaultUnaryOperatorNode_InstanceData_t));
    result = pInstanceData != NULL;
  }

  if (result)
  {
    pInstanceData->interfaceNodeUnaryOperatorNode.pDispatchTable =
      &gUnaryOperatorNodeDispatchTable;
    pInstanceData->interfaceNodeUnaryOperatorNode.pInstanceData =
      pInstanceData;

    pInstanceData->interfaceNodeScriptable.pDispatchTable =
      &gScriptableDispatchTable;
    pInstanceData->interfaceNodeScriptable.pInstanceData =
      pInstanceData;

    pInstanceData->referenceCount = 1;
    pInstanceData->operator = operator;

    /* We are about to make a copy of the operand node passed to this
     * constructor that we will keep as part of our instance data. As
     * a result, we need to add a reference to it.
     */
    Node_AddReference(pOperand);
    pInstanceData->pOperand = pOperand;

    /* Switch to the desired interface. This operation automatically
     * adds a reference to the reference it returns.
     */
    result = Fundamental_SwitchInterface(
      (Fundamental_t*)&(pInstanceData->interfaceNodeUnaryOperatorNode),
      pInterfaceName,
      ppResult);
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  /* Regardless of whether we have failed or succeeded in switching to
   * the desired interface, we need to remove the reference this
   * constructor has added to the instance (implicitly, by setting the
   * reference count to 1).
   */
  if (pInstanceData != NULL)
  {
    Fundamental_RemoveReference(
      (Fundamental_t*)&(pInstanceData->interfaceNodeUnaryOperatorNode));
  }

  return result;
}

static bool DefaultUnaryOperatorNode_SwitchInterface(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pInterfaceName,
  Fundamental_t** ppResult)
{
  bool result = (pInterfaceNode != NULL) && (pInterfaceName != NULL);
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if ((wcscmp(pInterfaceName, FUNDAMENTAL_NAME) == 0) ||
        (wcscmp(pInterfaceName, NODE_NAME) == 0) ||
        (wcscmp(pInterfaceName, UNARY_OPERATOR_NODE_NAME) == 0))
    {
      if (ppResult != NULL)
      {
        *ppResult =
          (Fundamental_t*)&(pThis->interfaceNodeUnaryOperatorNode);
      }
    }
    else if (wcscmp(pInterfaceName, SCRIPTABLE_NAME) == 0)
    {
      if (ppResult != NULL)
      {
        *ppResult = (Fundamental_t*)&(pThis->interfaceNodeScriptable);
      }
    }
    else
    {
      result = false;
    }
  }

  if (ppResult != NULL)
  {
    if (result)
    {
      Fundamental_AddReference(*ppResult);
    }
    else
    {
      *ppResult = NULL;
    }
  }

  return result;
}

static void DefaultUnaryOperatorNode_AddReference(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultUnaryOperatorNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;
    pThis->referenceCount++;
  }
}

static void DefaultUnaryOperatorNode_RemoveReference(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode)
{
  if (pInterfaceNode != NULL)
  {
    DefaultUnaryOperatorNode_InstanceData_t* pThis =
      pInterfaceNode->pInstanceData;

    pThis->referenceCount--;

    if (pThis->referenceCount == 0)
    {
      Node_RemoveReference(pThis->pOperand);
      free(pThis);
    }
  }
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_IS_CONSTANT (L"IsConstant")

static bool DefaultUnaryOperatorNode_IsConstant(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  bool* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;
  bool isOperandConstant = false;

  if (result)
  {
    result = Node_IsConstant(pThis->pOperand,
                             &isOperandConstant);
  }

  if (result)
  {
    *pResult = isOperandConstant;
  }
  else if (pResult != NULL)
  {
    *pResult = false;
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION                    \
  (L"PrintDebugInformation")

static bool DefaultUnaryOperatorNode_PrintDebugInformation(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  unsigned int startPosition,
  unsigned int indentationSize)
{
  bool result = pInterfaceNode != NULL;
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    if (startPosition != 0)
    {
      printf("%*s", startPosition, " ");
    }

    printf("DefaultUnaryOperatorNode[operator=%d, referenceCount=%d]\n",
           pThis->operator,
           pThis->referenceCount);

    result = Node_PrintDebugInformation(pThis->pOperand,
                                        startPosition + indentationSize,
                                        indentationSize);
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_OPERATOR (L"Operator")

static bool DefaultUnaryOperatorNode_Operator(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  UnaryOperator_t* pResult)
{
  bool result = (pInterfaceNode != NULL) && (pResult != NULL);
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    *pResult = pThis->operator;
  }
  else if (pResult != NULL)
  {
    *pResult = UnaryOperator_UNDEFINED;
  }

  return result;
}

/**
 * Make it possible to access this operation through the
 * <code>Scriptable</code> interface.
 */
#define SCRIPTABLE_OPERATION_OPERAND (L"Operand")

static bool DefaultUnaryOperatorNode_Operand(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  Node_t** ppResult)
{
  bool result = (pInterfaceNode != NULL) && (ppResult != NULL);
  DefaultUnaryOperatorNode_InstanceData_t* pThis =
    pInterfaceNode->pInstanceData;

  if (result)
  {
    Node_AddReference(pThis->pOperand);
    *ppResult = pThis->pOperand;
  }
  else if (ppResult != NULL)
  {
    *ppResult = NULL;
  }

  return result;
}

static bool DefaultUnaryOperatorNode_InvokeOperation(
  DefaultUnaryOperatorNode_InterfaceNode_t* pInterfaceNode,
  wchar_t* pOperationName,
  ScriptableReturnValue_t* pReturnValue,
  ScriptableArgumentType_t firstArgumentType,
  ...)
{
  bool result =
    (pInterfaceNode != NULL) &&
    (pOperationName != NULL) &&
    (pReturnValue != NULL) &&
    (firstArgumentType != ScriptableArgumentType_VOID);
  UnaryOperatorNode_t* pUnaryOperatorNode =
    (UnaryOperatorNode_t*)
    &(pInterfaceNode->pInstanceData->interfaceNodeUnaryOperatorNode);

  va_list arguments;
  va_start(arguments, firstArgumentType);

  if (result)
  {
    if (wcscmp(pOperationName, SCRIPTABLE_OPERATION_IS_CONSTANT) == 0)
    {
      bool isConstant = false;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = UnaryOperatorNode_IsConstant(pUnaryOperatorNode,
                                              &isConstant);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_BOOLEAN;
        pReturnValue->values.booleanValue = isConstant;
      }
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_PRINT_DEBUG_INFORMATION) == 0)
    {
      ScriptableArgumentType_t lastType = firstArgumentType;
      unsigned int argumentCount = 0;
      unsigned int startPosition = 0;
      unsigned int indentationSize = 0;

      // This operation takes two arguments, both integers.
      result = (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS);

      while (result &&
             (lastType != ScriptableArgumentType_NO_MORE_ARGUMENTS))
      {
        argumentCount++;
        result = argumentCount <= 2;

        if (result)
        {
          result = (lastType == ScriptableArgumentType_INTEGER);
        }

        if (result)
        {
          switch (argumentCount)
          {
            case 1:
              startPosition = va_arg(arguments, unsigned int);
              break;

            case 2:
              indentationSize = va_arg(arguments, unsigned int);
              break;

            default:
              result = false;
              break;
          }
        }

        if (result)
        {
          lastType = va_arg(arguments, ScriptableArgumentType_t);
        }
      }

      if (result)
      {
        result =
          UnaryOperatorNode_PrintDebugInformation(pUnaryOperatorNode,
                                                  startPosition,
                                                  indentationSize);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_VOID;
      }
    }
    else if (wcscmp(pOperationName, SCRIPTABLE_OPERATION_OPERATOR) == 0)
    {
      UnaryOperator_t operator = UnaryOperator_UNDEFINED;

      // This operation takes no input arguments.
      result = (firstArgumentType ==
                ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = UnaryOperatorNode_Operator(pUnaryOperatorNode,
                                            &operator);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_INTEGER;
        pReturnValue->values.integerValue = (unsigned int)operator;
      }
    }
    else if (wcscmp(pOperationName,
                    SCRIPTABLE_OPERATION_OPERAND) == 0)
    {
      Node_t* pNode = NULL;
      Scriptable_t* pScriptableNode = NULL;

      // This operation takes no input arguments.
      result =
        (firstArgumentType == ScriptableArgumentType_NO_MORE_ARGUMENTS);

      if (result)
      {
        result = UnaryOperatorNode_Operand(pUnaryOperatorNode,
                                           &pNode);
      }

      if (result)
      {
        /* The return value is an object accessed through the Node
         * interface. We can only return this value if the class
         * implementing this interface also supports the Scriptable
         * interface, and we thus need to test for this at runtime.
         */
        result = Node_SwitchInterface(pNode,
                                      SCRIPTABLE_NAME,
                                      (Fundamental_t**)&pScriptableNode);
      }

      if (result)
      {
        pReturnValue->type = ScriptableArgumentType_SCRIPTABLE;
        pReturnValue->values.pScriptableValue = pScriptableNode;
      }

      Node_RemoveReference(pNode);
    }
    else
    {
      result = false;
    }
  }

  va_end(arguments);

  if ((!result) && (pReturnValue != NULL))
  {
    pReturnValue->type = ScriptableArgumentType_INVALID;
  }

  return result;
}

/**
 * This is the dispatch table for the <code>UnaryOperatorNode</code>
 * interface implemented by this class.
 */
static const UnaryOperatorNode_DispatchTable_t
  gUnaryOperatorNodeDispatchTable =
{
  (bool (*)(UnaryOperatorNode_t*, wchar_t*, Fundamental_t**))
    DefaultUnaryOperatorNode_SwitchInterface,
  (void (*)(UnaryOperatorNode_t*))
    DefaultUnaryOperatorNode_AddReference,
  (void (*)(UnaryOperatorNode_t*))
    DefaultUnaryOperatorNode_RemoveReference,
  (bool (*)(UnaryOperatorNode_t*, bool*))
    DefaultUnaryOperatorNode_IsConstant,
  (bool (*)(UnaryOperatorNode_t*, unsigned int, unsigned int))
    DefaultUnaryOperatorNode_PrintDebugInformation,
  (bool (*)(UnaryOperatorNode_t*, UnaryOperator_t*))
    DefaultUnaryOperatorNode_Operator,
  (bool (*)(UnaryOperatorNode_t*, Node_t**))
    DefaultUnaryOperatorNode_Operand
};

/**
 * This is the dispatch table for the <code>Scriptable</code>
 * interface implemented by this class.
 */
static const Scriptable_DispatchTable_t gScriptableDispatchTable =
{
  (bool (*)(Scriptable_t*, wchar_t*, Fundamental_t**))
    DefaultUnaryOperatorNode_SwitchInterface,
  (void (*)(Scriptable_t*))
    DefaultUnaryOperatorNode_AddReference,
  (void (*)(Scriptable_t*))
    DefaultUnaryOperatorNode_RemoveReference,
  (bool (*)(Scriptable_t*,
            wchar_t*,
            ScriptableReturnValue_t*,
            ScriptableArgumentType_t,
            ...))
    DefaultUnaryOperatorNode_InvokeOperation
};
