/**
 * @file
 *
 * This file contains the interface <code>Fundamental</code>, which
 * all interfaces ultimately descend from. It contains operations that
 * facilitate interface navigation and reference counting.
 */

#ifndef INCLUSION_GUARD_FUNDAMENTAL
#define INCLUSION_GUARD_FUNDAMENTAL

#include <stdbool.h>
#include <wchar.h>

/**
 * The name of this interface.
 */
#define FUNDAMENTAL_NAME (L"se.polberger.components.Fundamental")

/**
 * Returns a reference to the underlying instance through a different
 * interface. <code>::AddReference()</code> is called automatically on
 * the new reference. If the class does not support the given
 * interface, this operation fails.
 *
 * @param[in] pFundamental
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[in] pInterfaceName
 *   the name of the interface. If this interface is not implemented
 *   by the underlying instance, this operation fails. Must not be
 *   <code>NULL</code>.
 * @param[out] ppResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, an interface reference. This parameter
 *   may be <code>NULL</code>, in which case this operation may be
 *   used to query an object as to whether it supports a given
 *   interface, with no concern for a result interface.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define Fundamental_SwitchInterface(pFundamental,                       \
                                    pInterfaceName,                     \
                                    ppResult)                           \
  (pFundamental)->pDispatchTable->SwitchInterface((pFundamental),       \
                                                  (pInterfaceName),     \
                                                  (ppResult));

/**
 * Notifies the instance implementing this interface that a new
 * reference has been added. This call always succeeds (thus there is
 * no return value).
 *
 * @param[in] pFundamental
 *   the instance implementing this interface. May be
 *   <code>NULL</code>.
 */
#define Fundamental_AddReference(pFundamental)                          \
  if ((pFundamental) != NULL)                                           \
  {                                                                     \
    (pFundamental)->pDispatchTable->AddReference(pFundamental);         \
  }

/**
 * Notifies the instance implementing this interface that a previously
 * added reference has been removed. If this instance finds that there
 * are no live references to it, it automatically destroys
 * itself. This call always succeeds (thus there is no return value).
 *
 * @param[in] pFundamental
 *   the instance implementing this interface. May be
 *   <code>NULL</code>.
 */
#define Fundamental_RemoveReference(pFundamental)                       \
  if ((pFundamental) != NULL)                                           \
  {                                                                     \
    (pFundamental)->pDispatchTable->RemoveReference(pFundamental);      \
  }

struct Fundamental_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct Fundamental_DispatchTable_s* pDispatchTable;
} Fundamental_t;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>Fundamental_t</code> type to match their own. (Only single
 * interface inheritance is supported.)
 */
typedef struct Fundamental_DispatchTable_s
{
  bool (*SwitchInterface)(Fundamental_t* pFundamental,
                          wchar_t* pInterfaceName,
                          Fundamental_t** ppResult);
  void (*AddReference)(Fundamental_t* pFundamental);
  void (*RemoveReference)(Fundamental_t* pFundamental);
} Fundamental_DispatchTable_t;

#endif // INCLUSION_GUARD_FUNDAMENTAL
