/**
 * @file
 *
 * This file contains an interface representing a literal operand node
 * in a syntax tree. This tree represents an arithmetic expression. A
 * literal operand node holds a constant (in the expression "3 / x",
 * "3" is a literal). This interface extends <code>Node</code>.
 */

#ifndef INCLUSION_GUARD_LITERAL_OPERAND_NODE
#define INCLUSION_GUARD_LITERAL_OPERAND_NODE

#include <stdbool.h>
#include "Fundamental.h"
#include "Node.h"

/**
 * The name of this interface.
 */
#define LITERAL_OPERAND_NODE_NAME                                       \
  (L"se.polberger.components.LiteralOperandNode")

/* Convenience macros for the operations inherited from the
 * Fundamental interface:
 */
#define LiteralOperandNode_SwitchInterface(pLiteralOperandNode,         \
                                           pInterfaceName,              \
                                           ppResult)                    \
  (pLiteralOperandNode)->pDispatchTable->SwitchInterface(               \
    (pLiteralOperandNode),                                              \
    (pInterfaceName),                                                   \
    (ppResult));
#define LiteralOperandNode_AddReference(pLiteralOperandNode)            \
  if ((pLiteralOperandNode) != NULL)                                    \
  {                                                                     \
    (pLiteralOperandNode)->pDispatchTable->AddReference(                \
      pLiteralOperandNode);                                             \
  }
#define LiteralOperandNode_RemoveReference(pLiteralOperandNode)         \
  if ((pLiteralOperandNode) != NULL)                                    \
  {                                                                     \
    (pLiteralOperandNode)->pDispatchTable->RemoveReference(             \
      pLiteralOperandNode);                                             \
  }

/* Convenience macros for the operations inherited from the Node
 * interface:
 */
#define LiteralOperandNode_IsConstant(pLiteralOperandNode, pResult)     \
  (pLiteralOperandNode)->pDispatchTable->IsConstant(                    \
    (pLiteralOperandNode), (pResult))
#define LiteralOperandNode_PrintDebugInformation(pLiteralOperandNode,   \
                                                 startPosition,         \
                                                 indentationSize)       \
  (pLiteralOperandNode)->pDispatchTable->PrintDebugInformation(         \
    (pLiteralOperandNode), (startPosition), (indentationSize))

// Convenience macros for the operations defined in this interface:

/**
 * Returns the constant of this node.
 *
 * @param[in] pLiteralOperandNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] ppResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the constant of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define LiteralOperandNode_Constant(pLiteralOperandNode, ppResult)      \
  (pLiteralOperandNode)->pDispatchTable->Constant((pLiteralOperandNode),\
                                                  (ppResult))

struct LiteralOperandNode_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct LiteralOperandNode_DispatchTable_s* pDispatchTable;
} LiteralOperandNode_t;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>LiteralOperandNode_t</code> type to match their own. (Only
 * single interface inheritance is supported.)
 */
typedef struct LiteralOperandNode_DispatchTable_s
{
  // Operations inherited from the Fundamental interface:
  bool (*SwitchInterface)(LiteralOperandNode_t* pLiteralOperandNode,
                          wchar_t* pInterfaceName,
                          Fundamental_t** ppResult);
  void (*AddReference)(LiteralOperandNode_t* pLiteralOperandNode);
  void (*RemoveReference)(LiteralOperandNode_t* pLiteralOperandNode);

  // Operations inherited from the Node interface:
  bool (*IsConstant)(LiteralOperandNode_t* pLiteralOperandNode,
                     bool* pResult);
  bool (*PrintDebugInformation)(
    LiteralOperandNode_t* pLiteralOperandNode,
    unsigned int startPosition,
    unsigned int indentationSize);

  // Operations defined in the LiteralOperandNode interface:
  bool (*Constant)(LiteralOperandNode_t* pLiteralOperandNode,
                   double* pResult);
} LiteralOperandNode_DispatchTable_t;

#endif // INCLUSION_GUARD_LITERAL_OPERAND_NODE
