/**
 * @file
 *
 * This file contains an interface, <code>Node</code>, representing an
 * immutable node in a syntax tree. This tree represents an arithmetic
 * expression as produced by an operator-precedence parser. This
 * interface is not meant to be implemented directly; rather, it
 * stipulates what functionality all nodes in said tree must be able
 * to provide. Descendant interfaces must add operations specific to
 * the type of node they are modeling.
 */

#ifndef INCLUSION_GUARD_NODE
#define INCLUSION_GUARD_NODE

#include <stdbool.h>
#include "Fundamental.h"

/**
 * The name of this interface.
 */
#define NODE_NAME (L"se.polberger.components.Node")

/* Convenience macros for the operations inherited from the
 * Fundamental interface:
 */
#define Node_SwitchInterface(pNode, pInterfaceName, ppResult)           \
  (pNode)->pDispatchTable->SwitchInterface((pNode),                     \
                                           (pInterfaceName),            \
                                           (ppResult));
#define Node_AddReference(pNode)                                        \
  if ((pNode) != NULL)                                                  \
  {                                                                     \
    (pNode)->pDispatchTable->AddReference(pNode);                       \
  }
#define Node_RemoveReference(pNode)                                     \
  if ((pNode) != NULL)                                                  \
  {                                                                     \
    (pNode)->pDispatchTable->RemoveReference(pNode);                    \
  }

// Convenience macros for the operations defined in this interface:

/**
 * Returns <code>true</code> if this node is fully constant,
 * <code>false</code> otherwise. Nodes which are fully constant may be
 * collapsed, meaning that they may be replaced with a literal node
 * representing the subtree of the node.
 *
 * @param[in] pNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] pResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, whether this node is constant. Must not
 *   be <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define Node_IsConstant(pNode, pResult)                                 \
  (pNode)->pDispatchTable->IsConstant((pNode), (pResult))

/**
 * Prints information on this node, suitable for debugging, to
 * standard output.
 *
 * @param[in] pNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define Node_PrintDebugInformation(pNode,                               \
                                   startPosition,                       \
                                   indentationSize)                     \
  (pNode)->pDispatchTable->PrintDebugInformation((pNode),               \
                                                 (startPosition),       \
                                                 (indentationSize))

struct Node_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct Node_DispatchTable_s* pDispatchTable;
} Node_t;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>Node_t</code> type to match their own. (Only single interface
 * inheritance is supported.)
 */
typedef struct Node_DispatchTable_s
{
  // Operations inherited from the Fundamental interface:
  bool (*SwitchInterface)(Node_t* pNode,
                          wchar_t* pInterfaceName,
                          Fundamental_t** ppResult);
  void (*AddReference)(Node_t* pNode);
  void (*RemoveReference)(Node_t* pNode);

  // Operations defined in the Node interface:
  bool (*IsConstant)(Node_t* pNode, bool* pResult);
  bool (*PrintDebugInformation)(Node_t* pNode,
                                unsigned int startPosition,
                                unsigned int indentationSize);
} Node_DispatchTable_t;

#endif // INCLUSION_GUARD_NODE
