/**
 * @file
 *
 * This file contains a program that tests the classes present in this
 * directory. In particular, it makes sure that dynamic dispatch works
 * as intended.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include "Node.h"
#include "BinaryOperatorNode.h"
#include "DefaultBinaryOperatorNode.h"
#include "UnaryOperatorNode.h"
#include "DefaultUnaryOperatorNode.h"
#include "LiteralOperandNode.h"
#include "DefaultLiteralOperandNode.h"
#include "IdentifierOperandNode.h"
#include "DefaultIdentifierOperandNode.h"
#include "Scriptable.h"

/**
 * Creates a test hierarchy of nodes, corresponding to this arithmetic
 * expression: ((-x - (6 * -3)) / z) + 2
 *
 * This is the resulting hierarchy:
 *
 * <code>
 *  1. BinaryOperatorNode[operator=+]
 *  2.   BinaryOperatorNode[operator=/]
 *  3.     BinaryOperatorNode[operator=-]
 *  4.       UnaryOperatorNode[operator=-]
 *  5.         IdentifierOperandNode[identifiercharacter=y]
 *  6.       BinaryOperatorNode[operator=*]
 *  7.         LiteralOperandNode[constant=6]
 *  8.         UnaryOperatorNode[operator=-]
 *  9.           LiteralOperandNode[constant=3]
 * 10.     IdentifierOperandNode[identifierCharacter=z]
 * 11.   LiteralOperandNode[constant=2]
 * </code>
 *
 * @param[out] ppRootNode
 *   a pointer to the variable which shall hold the instance of the
 *   root node (number 1, in the list above). Must not be
 *   <code>NULL</code>.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
static bool createNodes(Node_t** ppRootNode)
{
  bool result = (ppRootNode != NULL);

  /* The number suffixes of the variable names correspond to the
   * numbers given to the nodes in the list above.
   */
  Node_t* pNode1 = NULL;
  Node_t* pNode2 = NULL;
  Node_t* pNode3 = NULL;
  Node_t* pNode4 = NULL;
  Node_t* pNode5 = NULL;
  Node_t* pNode6 = NULL;
  Node_t* pNode7 = NULL;
  Node_t* pNode8 = NULL;
  Node_t* pNode9 = NULL;
  Node_t* pNode10 = NULL;
  Node_t* pNode11 = NULL;

  if (result)
  {
    result = DefaultLiteralOperandNode_Create(3,
                                              NODE_NAME,
                                              (Fundamental_t**)&pNode9);
  }

  if (result)
  {
    result = DefaultUnaryOperatorNode_Create(UnaryOperator_NEGATION,
                                             pNode9,
                                             NODE_NAME,
                                             (Fundamental_t**)&pNode8);
  }

  if (result)
  {
    result = DefaultLiteralOperandNode_Create(6,
                                              NODE_NAME,
                                              (Fundamental_t**)&pNode7);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_MULTIPLICATION,
                                              pNode7,
                                              pNode8,
                                              NODE_NAME,
                                              (Fundamental_t**)&pNode6);
  }

  if (result)
  {
    result = DefaultIdentifierOperandNode_Create(L'y',
                                                 NODE_NAME,
                                                 (Fundamental_t**)&pNode5);
  }

  if (result)
  {
    result = DefaultUnaryOperatorNode_Create(UnaryOperator_NEGATION,
                                             pNode5,
                                             NODE_NAME,
                                             (Fundamental_t**)&pNode4);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_SUBTRACTION,
                                              pNode4,
                                              pNode6,
                                              NODE_NAME,
                                              (Fundamental_t**)&pNode3);
  }

  if (result)
  {
    result = DefaultIdentifierOperandNode_Create(L'z',
                                                 NODE_NAME,
                                                 (Fundamental_t**)&pNode10);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_DIVISION,
                                              pNode3,
                                              pNode10,
                                              NODE_NAME,
                                              (Fundamental_t**)&pNode2);
  }

  if (result)
  {
    result = DefaultLiteralOperandNode_Create(2,
                                              NODE_NAME,
                                              (Fundamental_t**)&pNode11);
  }

  if (result)
  {
    result = DefaultBinaryOperatorNode_Create(BinaryOperator_ADDITION,
                                              pNode2,
                                              pNode11,
                                              NODE_NAME,
                                              (Fundamental_t**)&pNode1);
  }

  if (result)
  {
    *ppRootNode = pNode1;
  }
  else
  {
    *ppRootNode = NULL;
  }

  /* Remove our references to all nodes but pNode1, which is the node
   * we are returning.
   */
  Node_RemoveReference(pNode11);
  Node_RemoveReference(pNode10);
  Node_RemoveReference(pNode9);
  Node_RemoveReference(pNode8);
  Node_RemoveReference(pNode7);
  Node_RemoveReference(pNode6);
  Node_RemoveReference(pNode5);
  Node_RemoveReference(pNode4);
  Node_RemoveReference(pNode3);
  Node_RemoveReference(pNode2);

  return result;
}

/**
 * Prints some information on the root node given as a parameter to
 * this function, as well as some related to its operands. This
 * implementation uses late binding -- in other words, the
 * implementation is bound to dynamically, but invocations are checked
 * statically (types and parameters).
 *
 * @param[in] pRootNodeAsNode
 *   a Node reference to the root node of the tree.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
bool experimentWithLateBinding(Node_t* pRootNodeAsNode)
{
  BinaryOperatorNode_t* pRootNode = NULL;
  Node_t* pLeftOperand = NULL;
  Node_t* pRightOperand = NULL;
  bool rootIsConstant = false;
  bool leftOperandIsConstant = false;
  bool rightOperandIsConstant = false;
  bool result = pRootNodeAsNode != NULL;

  if (result)
  {
    result = Node_SwitchInterface(pRootNodeAsNode,
                                  BINARY_OPERATOR_NODE_NAME,
                                  (Fundamental_t**)&pRootNode);
  }

  if (result)
  {
    result = BinaryOperatorNode_LeftOperand(pRootNode,
                                            &pLeftOperand);
  }

  if (result)
  {
    result = BinaryOperatorNode_RightOperand(pRootNode,
                                             &pRightOperand);
  }

  if (result)
  {
    result = Node_IsConstant(pLeftOperand,
                             &leftOperandIsConstant);
  }
  
  if (result)
  {
    result = Node_IsConstant(pRightOperand,
                             &rightOperandIsConstant);
  }

  if (result)
  {
    printf("Root is constant: %d.\n", rootIsConstant);
    printf("Left operand is constant: %d.\n", leftOperandIsConstant);
    printf("Right operand is constant: %d.\n", rightOperandIsConstant);
  }

  Node_RemoveReference(pRightOperand);
  Node_RemoveReference(pLeftOperand);
  BinaryOperatorNode_RemoveReference(pRootNode);

  return result;
}

/**
 * Prints some information on the root node given as a parameter to
 * this function, as well as some related to its operands. This
 * implementation uses very late binding -- in other words, everything
 * is done dynamically: finding the proper implementation, as well as
 * checking the validity of the invocations themselves.
 *
 * @param[in] pRootNode
 *   a Node reference to the root node of the tree.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
bool experimentWithVeryLateBinding(Node_t* pRootNode)
{
  Scriptable_t* pScriptableRootNode = NULL;
  ScriptableReturnValue_t returnValueLeftOperand;
  ScriptableReturnValue_t returnValueRightOperand;
  ScriptableReturnValue_t returnValueRootIsConstant;
  ScriptableReturnValue_t returnValueLeftOperandIsConstant;
  ScriptableReturnValue_t returnValueRightOperandIsConstant;
  bool result = pRootNode != NULL;

  if (result)
  {
    result = Node_SwitchInterface(pRootNode,
                                  SCRIPTABLE_NAME,
                                  (Fundamental_t**)&pScriptableRootNode);
  }

  if (result)
  {
    result = 
      Scriptable_InvokeOperation(pScriptableRootNode,
                                 L"LeftOperand",
                                 &returnValueLeftOperand,
                                 ScriptableArgumentType_NO_MORE_ARGUMENTS);
  }

  if (result)
  {
    result = (returnValueLeftOperand.type == ScriptableArgumentType_SCRIPTABLE);
  }

  if (result)
  {
    result =
      Scriptable_InvokeOperation(pScriptableRootNode,
                                 L"RightOperand",
                                 &returnValueRightOperand,
                                 ScriptableArgumentType_NO_MORE_ARGUMENTS);
  }

  if (result)
  {
    result = 
      (returnValueRightOperand.type == ScriptableArgumentType_SCRIPTABLE);
  }

  if (result)
  {
    result =
      Scriptable_InvokeOperation(returnValueLeftOperand.values.pScriptableValue,
                                 L"IsConstant",
                                 &returnValueLeftOperandIsConstant,
                                 ScriptableArgumentType_NO_MORE_ARGUMENTS);
  }

  if (result)
  {
    result = (returnValueLeftOperandIsConstant.type == ScriptableArgumentType_BOOLEAN);
  }

  if (result)
  {
    result =
      Scriptable_InvokeOperation(
        returnValueRightOperand.values.pScriptableValue,
        L"IsConstant",
        &returnValueRightOperandIsConstant,
        ScriptableArgumentType_NO_MORE_ARGUMENTS);
  }

  if (result)
  {
    result = (returnValueRightOperandIsConstant.type == 
              ScriptableArgumentType_BOOLEAN);
  }

  if (result)
  {
    result =
      Scriptable_InvokeOperation(pScriptableRootNode,
                                 L"IsConstant",
                                 &returnValueRootIsConstant,
                                 ScriptableArgumentType_NO_MORE_ARGUMENTS);
  }

  if (result)
  {
    result = (returnValueRootIsConstant.type == ScriptableArgumentType_BOOLEAN);
  }

  if (result)
  {
    printf("Root is constant: %d.\n",
           returnValueRootIsConstant.values.booleanValue);
    printf("Left operand is constant: %d.\n",
           returnValueLeftOperandIsConstant.values.booleanValue);
    printf("Right operand is constant: %d.\n",
           returnValueRightOperandIsConstant.values.booleanValue);
  }

  if (returnValueLeftOperand.type == ScriptableArgumentType_SCRIPTABLE)
  {
    Scriptable_RemoveReference(returnValueLeftOperand.values.pScriptableValue);
  }

  if (returnValueRightOperand.type == ScriptableArgumentType_SCRIPTABLE)
  {
    Scriptable_RemoveReference(returnValueRightOperand.values.pScriptableValue);
  }

  Scriptable_RemoveReference(pScriptableRootNode);

  return result;
}

/**
 * The main function of this program. It builds a sample arithmetic
 * syntax tree, and prints some information on it using both late
 * binding and very late binding. Essentially, this is a testbed for
 * the object-oriented framework developed in the directory in which
 * this file resides.
 *
 * @return
 *   <code>0</code> if successful, a different value otherwise.
 */
int main()
{
  bool result = true;
  Node_t* pRootNode = NULL;

  result = createNodes(&pRootNode);

  if (result)
  {
    printf("Experimenting with late binding:\n");
    result = experimentWithLateBinding(pRootNode);
 }

  if (result)
  {
    printf("\n"); 
    printf("Experimenting with very late binding:\n");
    result = experimentWithVeryLateBinding(pRootNode);
  }

  if (result)
  {
    printf("\n"); 
    printf("Tree (using late binding):\n");
    result = Node_PrintDebugInformation(pRootNode, 0, 2);
  }

  if (result)
  {
    printf("\n"); 
    printf("Tree (using very late binding):\n");

    Scriptable_t* pScriptableRootNode = NULL;
    ScriptableReturnValue_t returnValue;

    result = 
      Scriptable_SwitchInterface(pRootNode,
                                 SCRIPTABLE_NAME,
                                 (Fundamental_t**)&pScriptableRootNode);

    if (result)
    {
      result = Scriptable_InvokeOperation(
        pScriptableRootNode,
        L"PrintDebugInformation",
        &returnValue,
        ScriptableArgumentType_INTEGER,
        0,
        ScriptableArgumentType_INTEGER,
        2,
        ScriptableArgumentType_NO_MORE_ARGUMENTS);
    }

    if (result)
    {
      result = (returnValue.type == ScriptableArgumentType_VOID);
    }

    /* While we don't expect the return type to be of type Scriptable,
     * we are contractually obligated to remove the reference if this
     * type is returned.
     */
    if (returnValue.type == ScriptableArgumentType_SCRIPTABLE)
    {
      Scriptable_RemoveReference(returnValue.values.pScriptableValue);
    }

    Scriptable_RemoveReference(pScriptableRootNode);
  }

  if (!result)
  {
    printf("Something did not go as intended.\n");
  }

  Node_RemoveReference(pRootNode);

  return result ? 0 : 1;
}
