/**
 * @file
 *
 * This file contains an interface, <code>Scriptable</code>, which
 * allows invocations to be checked at runtime. In other words, it
 * facilitates very late binding. With late binding, the
 * implementations of interfaces are found at runtime, but invocations
 * are checked statically. An object implementing this interface may
 * be called from environments that do not allow invocations to be
 * checked statically, such as scripting languages.
 */

#ifndef INCLUSION_GUARD_SCRIPTABLE
#define INCLUSION_GUARD_SCRIPTABLE

#include <stdbool.h>
#include <wchar.h>
#include "Fundamental.h"
#include "ScriptableArgumentType.h"

/**
 * The name of this interface.
 */
#define SCRIPTABLE_NAME (L"se.polberger.components.Scriptable")

/* Convenience macros for the operations inherited from the
 * Fundamental interface:
 */
#define Scriptable_SwitchInterface(pScriptable,                         \
                                   pInterfaceName,                      \
                                   ppResult)                            \
  (pScriptable)->pDispatchTable->SwitchInterface(pScriptable,           \
                                                 pInterfaceName,        \
                                                 ppResult);
#define Scriptable_AddReference(pScriptable)                            \
  if (pScriptable != NULL)                                              \
  {                                                                     \
    (pScriptable)->pDispatchTable->AddReference(pScriptable);           \
  }
#define Scriptable_RemoveReference(pScriptable)                         \
  if (pScriptable != NULL)                                              \
  {                                                                     \
    (pScriptable)->pDispatchTable->RemoveReference(pScriptable);        \
  }

// Convenience macros for the operations defined in this interface:

/**
 * Invokes the specified operation with the given arguments. This
 * operation will fail if the underlying implementation does not
 * support the given operation, or if the argument list is not
 * correct. All arguments must be of the correct types.
 *
 * The <code>firstArgumentType</code> must be set to the type of the
 * first argument, or
 * <code>ScriptableArgumentType_NO_MORE_ARGUMENTS</code> if the called
 * operation accepts no arguments. It is followed by the data of the
 * first argument. If there are additional arguments, the next
 * argument must be set to the type of the next argument, followed by
 * its data, and so on. The last given argument to this operation must
 * be <code>ScriptableArgumentType_NO_MORE_ARGUMENTS</code>.
 *
 * @param[in] pScriptable
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[in] pOperationName
 *   the name of the operation. This string is case-sensitive. Must
 *   not be <code>NULL</code.>
 * @param[out] pReturnValue
 *   a pointer to the variable which shall hold the return value
 *   returned from the called operation. This variable is normally
 *   allocated on the stack. If the return value holds a reference to
 *   an object implementing the <code>Scriptable</code> interface
 *   (that is, if <code>pReturnValue->type ==
 *   ScriptableArgumentType_SCRIPTABLE</code>), the caller is
 *   responsible for removing the reference that has been added.
 * @param[in] firstArgumentType
 *   the type of the first argument. This must be set to
 *   <code>ScriptableArgumentType_NO_MORE_ARGUMENTS</code> if the
 *   called operation does not accept any arguments.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define Scriptable_InvokeOperation(pScriptable,                         \
                                   pOperationName,                      \
                                   pReturnValue,                        \
                                   ...)                                 \
  (pScriptable)->pDispatchTable->InvokeOperation(pScriptable,           \
                                                 pOperationName,        \
                                                 pReturnValue,          \
                                                 __VA_ARGS__)

struct Scriptable_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct Scriptable_DispatchTable_s* pDispatchTable;
} Scriptable_t;

struct ScriptableReturnValue_s;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>Scriptable_t</code> type to match their own. (Only single
 * interface inheritance is supported.)
 */
typedef struct Scriptable_DispatchTable_s
{
  // Operations inherited from the Fundamental interface:
  bool (*SwitchInterface)(Scriptable_t* pThis,
                          wchar_t* pInterfaceName,
                          Fundamental_t** ppResult);
  void (*AddReference)(Scriptable_t* pThis);
  void (*RemoveReference)(Scriptable_t* pThis);

  // Operations defined in the Scriptable interface:
  bool (*InvokeOperation)(Scriptable_t* pThis,
                          wchar_t* pOperationName,
                          struct ScriptableReturnValue_s* pReturnValue,
                          ScriptableArgumentType_t firstArgumentType,
                          ...);
} Scriptable_DispatchTable_t;

/**
 * This type represents a return value from a scriptable operation. It
 * is a tagged union, that is, a structure which contains the type of
 * its data, as well as a union representing the data.
 */
typedef struct ScriptableReturnValue_s
{
  /**
   * The type of the data stored in this tagged union (its
   * <em>tag</em>).
   */
  ScriptableArgumentType_t type;

  union
  {
    /**
     * Data corresponding to
     * <code>ScriptableArgumentType_INTEGER</code>.
     */
    unsigned int integerValue;

    /**
     * Data corresponding to
     * <code>ScriptableArgumentType_DOUBLE</code>.
     */
    double doubleValue;

    /**
     * Data corresponding to
     * <code>ScriptableArgumentType_BOOLEAN</code>.
     */
    bool booleanValue;

    /**
     * Data corresponding to
     * <code>ScriptableArgumentType_CHARACTER</code>.
     */
    wchar_t characterValue;

    /**
     * Data corresponding to
     * <code>ScriptableArgumentType_SCRIPTABLE</code>. If the return
     * value contains data of this type, the reference must be removed
     * at some point.
     */
    Scriptable_t* pScriptableValue;
  } values;
} ScriptableReturnValue_t;

#endif // INCLUSION_GUARD_SCRIPTABLE
