/**
 * @file
 *
 * This file contains an interface, <code>UnaryOperatorNode</code>,
 * representing a unary operator node in a syntax tree. This tree
 * represents an arithmetic expression. A unary operator takes one
 * operand (in the expression "3 + -2", the negation operator operates
 * on the literal operand "2"). This interface extends
 * <code>Node</code>.
 */

#ifndef INCLUSION_GUARD_UNARY_OPERATOR_NODE
#define INCLUSION_GUARD_UNARY_OPERATOR_NODE

#include <stdbool.h>
#include "Fundamental.h"
#include "Node.h"
#include "UnaryOperator.h"

/**
 * The name of this interface.
 */
#define UNARY_OPERATOR_NODE_NAME                                        \
  (L"se.polberger.components.UnaryOperatorNode")

/* Convenience macros for the operations inherited from the
 * Fundamental interface:
 */
#define UnaryOperatorNode_SwitchInterface(pUnaryOperatorNode,           \
                                          pInterfaceName,               \
                                          ppResult)                     \
  (pUnaryOperatorNode)->pDispatchTable->SwitchInterface(                \
    (pUnaryOperatorNode),                                               \
    (pInterfaceName),                                                   \
    (ppResult));
#define UnaryOperatorNode_AddReference(pUnaryOperatorNode)              \
  if ((pUnaryOperatorNode) != NULL)                                     \
  {                                                                     \
    (pUnaryOperatorNode)->pDispatchTable->AddReference(                 \
      pUnaryOperatorNode);                                              \
  }
#define UnaryOperatorNode_RemoveReference(pUnaryOperatorNode)           \
  if ((pUnaryOperatorNode) != NULL)                                     \
  {                                                                     \
    (pUnaryOperatorNode)->pDispatchTable->RemoveReference(              \
      pUnaryOperatorNode);                                              \
  }

/* Convenience macros for the operations inherited from the Node
 * interface:
 */
#define UnaryOperatorNode_IsConstant(pUnaryOperatorNode, pResult)       \
  (pUnaryOperatorNode)->pDispatchTable->IsConstant((pUnaryOperatorNode),\
                                                   (pResult))
#define UnaryOperatorNode_PrintDebugInformation(pUnaryOperatorNode,     \
                                                startPosition,          \
                                                indentationSize)        \
  (pUnaryOperatorNode)->pDispatchTable->PrintDebugInformation(          \
    (pUnaryOperatorNode), (startPosition), (indentationSize))

// Convenience macros for the operations defined in this interface:

/**
 * Returns the unary operator of this node.
 *
 * @param[in] pUnaryOperatorNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] pResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the unary operator of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define UnaryOperatorNode_Operator(pUnaryOperatorNode, pResult)         \
  (pUnaryOperatorNode)->pDispatchTable->Operator((pUnaryOperatorNode),  \
    (pResult))

/**
 * Returns the sole operand of this node.
 *
 * @param[in] pUnaryOperatorNode
 *   the instance implementing this interface. Must not be
 *   <code>NULL</code>.
 * @param[out] ppResult
 *   a pointer to the variable which shall hold the result returned by
 *   this operation, that is, the sole operand of this node.
 * @return
 *   <code>true</code> if the operation completes successfully,
 *   <code>false</code> otherwise.
 */
#define UnaryOperatorNode_Operand(pUnaryOperatorNode, ppResult)         \
  (pUnaryOperatorNode)->pDispatchTable->Operand((pUnaryOperatorNode),   \
                                                (ppResult))

struct UnaryOperatorNode_DispatchTable_s;

/**
 * Variables of this type may be used to represent objects
 * implementing this interface.
 */
typedef struct
{
  const struct UnaryOperatorNode_DispatchTable_s* pDispatchTable;
} UnaryOperatorNode_t;

/**
 * This is the dispatch table, or vtable, for this interface. It
 * represents an indirection that enables the implementation of an
 * interface to be bound to at runtime. Interfaces wishing to extend
 * this interface must include the complete contents of this structure
 * as the first members of their dispatch tables, changing the
 * <code>UnaryOperatorNode_t</code> type to match their own. (Only
 * single interface inheritance is supported.)
 */
typedef struct UnaryOperatorNode_DispatchTable_s
{
  // Operations inherited from the Fundamental interface:
  bool (*SwitchInterface)(UnaryOperatorNode_t* pUnaryOperatorNode,
                          wchar_t* pInterfaceName,
                          Fundamental_t** ppResult);
  void (*AddReference)(UnaryOperatorNode_t* pUnaryOperatorNode);
  void (*RemoveReference)(UnaryOperatorNode_t* pUnaryOperatorNode);

  // Operations inherited from the Node interface:
  bool (*IsConstant)(UnaryOperatorNode_t* pUnaryOperatorNode,
                     bool* pResult);
  bool (*PrintDebugInformation)(UnaryOperatorNode_t* pUnaryOperatorNode,
                                unsigned int startPosition,
                                unsigned int indentationSize);

  // Operations defined in the UnaryOperatorNode interface:
  bool (*Operator)(UnaryOperatorNode_t* pUnaryOperatorNode,
                   UnaryOperator_t* pResult);
  bool (*Operand)(UnaryOperatorNode_t* pUnaryOperatorNode,
                  Node_t** ppResult);
} UnaryOperatorNode_DispatchTable_t;

#endif // INCLUSION_GUARD_UNARY_OPERATOR_NODE
